from __future__ import annotations

import asyncio
from asyncio import Event
from copy import copy
from dataclasses import dataclass, field
from math import sqrt
from typing import Awaitable, Generic, Self

import numpy as np

from ..async_utils import asyncio_run
from ..states import State, TMove
from ..states.state import Policy
from .network import NeuralNetwork


@dataclass
class AlphaZeroMcts:
    neural_net: NeuralNetwork
    num_mcts_sims: int
    dirichlet_epsilon: float
    dirichlet_alpha: float
    discount_factor: float

    def policy(self, state: State[TMove], root_node: Node[TMove] | None = None) -> Policy[TMove]:
        return asyncio_run(self.policy_async(state, root_node))

    async def policy_async(
        self,
        state: State[TMove],
        root_node: Node[TMove] | None = None,
        cancellation_event: Event | None = None,
    ) -> Awaitable[Policy[TMove]]:
        root_node = await self.search_async(state, root_node, cancellation_event)
        root_q = 1 - root_node.q_value
        value = root_q * 2 - 1

        moves = [node.move for node in root_node.children]
        priors = np.array([node.visit_count for node in root_node.children], dtype=np.float32)
        priors /= priors.sum()

        policy_size = self.neural_net.game.config().action_size
        encoded_policy = np.zeros(policy_size)
        for move, prior in zip(moves, priors):
            loc = state.policy_loc(move)
            encoded_policy[loc] = prior

        return Policy(moves, priors, encoded_policy, value, root_node)

    def policies(
        self, states: list[State[TMove]], root_nodes: list[Node[TMove]] | None = None
    ) -> list[Policy]:
        root_nodes = self.search_parallel(states, root_nodes)

        policies: list[Policy] = []
        policy_shape = self.neural_net.game.config().action_size

        for state, root_node in zip(states, root_nodes):
            root_q = 1 - root_node.q_value
            value = root_q * 2 - 1

            moves = [node.move for node in root_node.children]
            new_priors = np.array([node.visit_count for node in root_node.children], dtype=np.float32)
            new_priors /= new_priors.sum()

            encoded_policy = np.zeros(policy_shape)
            for move, prior in zip(moves, new_priors):
                loc = state.policy_loc(move)
                encoded_policy[loc] = prior

            policy = Policy(moves, new_priors, encoded_policy, value, root_node)
            policies.append(policy)

        return policies

    def search(self, root_state: State[TMove], root_node: Node[TMove] | None = None) -> Node[TMove]:
        return asyncio_run(self.search_async(root_state, root_node))

    async def search_async(
        self,
        root_state: State[TMove],
        root_node: Node[TMove] | None = None,
        cancellation_event: Event | None = None,
    ) -> Awaitable[Node[TMove]]:
        root: Node[TMove] = root_node or Node(root_state.played_by)

        if root_node:
            # Add Dirchlet noise to the children if they are present
            ε = self.dirichlet_epsilon
            alpha = self.dirichlet_alpha
            noise = np.random.dirichlet([alpha] * len(root.children))
            for i, child_node in enumerate(root.children):
                child_node.prior = (1 - ε) * child_node.prior + ε * noise[i]

            root.parent = None  # free memory

        # check to bail more frequently if we are pondering - keep responsive
        poll_rate = 1 if cancellation_event else 5

        for n in range(self.num_mcts_sims):
            if n % poll_rate == 0:
                await asyncio.sleep(0)
            if cancellation_event and cancellation_event.is_set():
                break

            node = root
            state = copy(root_state)

            # === Selection ===
            while node.has_children:
                node = node.select_child()
                state.set_move(node.move)

            status = state.status()
            if status.is_in_progress:
                raw_policy, value = self.neural_net.predict(state)

                # Get the priors
                priors = np.zeros(len(status.legal_moves))
                for i, move in enumerate(status.legal_moves):
                    loc = state.policy_loc(move)
                    priors[i] = raw_policy[loc]
                priors /= np.sum(priors)

                if n == 0 and root_node is None:
                    ε = self.dirichlet_epsilon
                    alpha = self.dirichlet_alpha
                    noise = np.random.dirichlet([alpha] * len(priors))
                    priors = (1 - ε) * priors + ε * noise

                # === Expansion ===
                for move, prior in zip(status.legal_moves, priors):
                    child_played_by = 2 if state.played_by == 1 else 1
                    child_node = Node(child_played_by, parent=node, move=move, prior=prior)
                    node.children.append(child_node)

                # === Simulation ===
                # Here, the AlphaZero paper completely replaces the traditional rollout phase with
                # a value estimation from the neural net.
                # Negate because the net gives an estimation from player whose turn it is next,
                # rather than the player who has just moved
                value *= -1
            else:
                value = status.value

            # === Backpropagate ===
            node.backpropagate(value, self.discount_factor)

        return root

    def search_parallel(
        self, root_states: list[State[TMove]], root_nodes: list[Node[TMove]] | None = None
    ) -> list[Node]:
        roots = root_nodes or [Node(root_state.played_by) for root_state in root_states]
        if root_nodes:
            for root in roots:
                # Add Dirchlet noise to the children if they are present
                ε = self.dirichlet_epsilon
                alpha = self.dirichlet_alpha
                noise = np.random.dirichlet([alpha] * len(root.children))
                for i, child_node in enumerate(root.children):
                    child_node.prior = (1 - ε) * child_node.prior + ε * noise[i]

                root.parent = None  # free memory

        for n in range(self.num_mcts_sims):
            nodes = [root for root in roots]
            states: list[State[TMove]] = [copy(root_state) for root_state in root_states]

            # === Selection ===
            for j in range(len(nodes)):
                while nodes[j].has_children:
                    nodes[j] = nodes[j].select_child()
                    states[j].set_move(nodes[j].move)

            statuses = [state.status() for state in states]
            in_progress_idxs = [i for i, status in enumerate(statuses) if status.is_in_progress]
            finished_idxs = [i for i, status in enumerate(statuses) if not status.is_in_progress]
            num_in_progress = len(in_progress_idxs)
            values = np.zeros(len(statuses), dtype=np.float32)

            if num_in_progress > 0:
                in_progress_states = [states[idx] for idx in in_progress_idxs]
                raw_policies, predicted_values = self.neural_net.predict_parallel(in_progress_states)

                for i, idx in enumerate(in_progress_idxs):
                    state = states[idx]
                    status = statuses[idx]
                    raw_policy = raw_policies[i]
                    legal_moves = status.legal_moves

                    # Get the priors
                    priors = np.zeros(len(legal_moves))
                    for j, move in enumerate(legal_moves):
                        loc = state.policy_loc(move)
                        priors[j] = raw_policy[loc]
                    priors /= np.sum(priors)

                    if n == 0 and root_nodes is None:
                        ε = self.dirichlet_epsilon
                        alpha = self.dirichlet_alpha
                        noise = np.random.dirichlet([alpha] * len(priors))
                        priors = (1 - ε) * priors + ε * noise

                    # === Expansion ===
                    for move, prior in zip(legal_moves, priors):
                        child_played_by = 2 if state.played_by == 1 else 1
                        child_node = Node(child_played_by, nodes[idx], move, prior)
                        nodes[idx].children.append(child_node)

                # === Simulation ===
                # Here, the AlphaZero paper completely replaces the traditional rollout phase with
                # a value estimation from the neural net.
                # Negate because the net gives an estimation from player whose turn it is next,
                # rather than the player who has just moved
                values[in_progress_idxs] = -1 * predicted_values

            if finished_idxs:
                values[finished_idxs] = [statuses[idx].value for idx in finished_idxs]

            # === Backpropagate ===
            for i, node in enumerate(nodes):
                node.backpropagate(values[i], self.discount_factor)

        return roots


@dataclass(slots=True)
class Node(Generic[TMove]):
    played_by: int
    parent: Node[TMove] | None = None
    move: TMove | None = None
    prior: np.float32 = 1.0

    value_sum: float = field(default_factory=float, init=False)
    visit_count: int = field(default_factory=int, init=False)
    children: list[Node[TMove]] = field(default_factory=list, init=False)

    @property
    def has_children(self) -> bool:
        return bool(self.children)

    @property
    def q_value(self) -> float:
        if self.visit_count == 0:
            if self.parent is None:
                return 0

            FIRST_PLAY_URGENCY = 0.1  # 0.44
            q_from_parent = 1 - self.parent.q_value
            estimated_q_value = q_from_parent - FIRST_PLAY_URGENCY
            return max(estimated_q_value, 0)

        return (1 + self.value_sum / self.visit_count) / 2

    def update(self, outcome: float) -> None:
        self.visit_count += 1
        self.value_sum += outcome

    def backpropagate(self, outcome: float, discount_factor: float) -> None:
        self.update(outcome)
        if self.parent:
            self.parent.backpropagate(-outcome * discount_factor, discount_factor)

    def select_child(self) -> Self:
        return max(self.children, key=lambda c: c.ucb())

    def ucb(self) -> float:
        c = 2  # todo AlphaZero sets to.... 2?
        exploration_param = sqrt(self.parent.visit_count) / (1 + self.visit_count)
        return self.q_value + c * self.prior * exploration_param

    def __repr__(self):
        if self.parent:
            return f"Node(move={self.move}, Q={self.q_value:.1%}, prior={self.prior:.2%}, visit_count={self.visit_count}, UCB={self.ucb():.3})"
        return f"Node(move={self.move}, Q={self.q_value:.1%}, prior={self.prior:.2%}, visit_count={self.visit_count})"



from __future__ import annotations

from asyncio import Event
from copy import deepcopy
from dataclasses import asdict, dataclass
from itertools import product
from typing import Awaitable, cast

import numpy as np
import tqdm
from torch.utils.tensorboard import SummaryWriter
from tqdm import trange

from ..games import Game
from ..solvers import Solver
from ..states import State, TMove
from ..states.state import Policy, TemperatureSchedule
from .alpha_zero_mcts import AlphaZeroMcts, Node
from .alpha_zero_parameters import AZArenaParameters, AZMctsParameters, AZTrainingParameters
from .alpha_zero_training_models import ParallelGame, RecordedAction
from .network import NeuralNetwork, TrainingData


class AlphaZero:
    def __init__(self, neural_net: NeuralNetwork) -> None:
        self.neural_net = neural_net

    @property
    def game(self) -> Game:
        return self.neural_net.game

    def as_solver(self, num_mcts_sims: int | AZMctsParameters) -> Solver:
        return AlphaZeroSolver(self, num_mcts_sims)

    def policy(
        self,
        state: State[TMove],
        num_mcts_sims: int | AZMctsParameters,
        root_node: Node[TMove] | None = None,
    ) -> Policy:
        self.neural_net.set_to_eval()
        mcts = self._get_mcts(num_mcts_sims)
        return mcts.policy(state, root_node)

    def policy_async(
        self,
        state: State[TMove],
        num_mcts_sims: int | AZMctsParameters,
        root_node: Node[TMove] | None = None,
        cancellation_event: Event | None = None,
    ) -> Awaitable[Policy]:
        self.neural_net.set_to_eval()
        mcts = self._get_mcts(num_mcts_sims)
        return mcts.policy_async(state, root_node, cancellation_event)

    def policies(
        self, states: list[State[TMove]], num_mcts_sims: int | AZMctsParameters
    ) -> list[Policy]:
        self.neural_net.set_to_eval()
        mcts = self._get_mcts(num_mcts_sims)
        return mcts.policies(states)

    def raw_policy(self, state: State[TMove]) -> Policy:
        self.neural_net.set_to_eval()
        encoded_policy, value = self.neural_net.predict(state)
        return Policy(None, None, encoded_policy, value)

    def select_move(
        self,
        state: State[TMove],
        num_mcts_sims: int | AZMctsParameters,
        temperature_schedule: TemperatureSchedule | None = None,
    ) -> TMove:
        schedule = temperature_schedule or TemperatureSchedule.competitive()
        temperature = schedule.get_temperature(state.move_count)
        policy = self.policy(state, num_mcts_sims)
        move = policy.select_move(temperature)
        return move

    def select_moves(
        self,
        states: list[State[TMove]],
        num_mcts_sims: int | AZMctsParameters,
        temperature_schedule: TemperatureSchedule | None = None,
    ) -> list[TMove]:
        schedule = temperature_schedule or TemperatureSchedule.competitive()
        temperatures = [schedule.get_temperature(state.move_count) for state in states]
        policies = self.policies(states, num_mcts_sims)
        moves = [policy.select_move(temperature) for policy, temperature in zip(policies, temperatures)]
        return moves

    def self_learn(
        self, training_params: AZTrainingParameters, initial_state: str | list[int] | None = None
    ) -> None:
        num_rounds = training_params.games_per_generation // training_params.num_parallel

        mcts = self._get_mcts(training_params.mcts_parameters)

        for _ in trange(training_params.num_generations, desc="Generations"):
            training_set: list[TrainingData] = []

            for _ in trange(num_rounds, desc="- Self-play", leave=False):
                # training_set += self.self_play(mcts, training_params.temperature, initial_state)
                training_set += self.self_play_parallel(
                    mcts,
                    training_params.num_parallel,
                    training_params.temperature,
                    initial_state,
                )

            extended_training_set = self._exploit_symmetries(training_set)

            broken = """

            neural_net_after = deepcopy(self.neural_net)
            neural_net_after.train(extended_training_set, num_epochs, minibatch_size)
            neural_net_after.generation += 1

            # Battle the newly trained net in the arena
            arena = Arena(self.game, **training_params.arena_parameters.__dict__)
            alpha_zero_after = AlphaZero(neural_net_after)
            arena_result = arena.battle(self, alpha_zero_after)
            if arena_result.is_improvement or not arena_result.must_improve:
                self.neural_net = neural_net_after
                self.neural_net.save()
            """

            # Train against the newly generated games
            num_epochs, minibatch_size = training_params.num_epochs, training_params.minibatch_size
            self.neural_net.save_training_data(extended_training_set)
            neural_net_before = deepcopy(self.neural_net)
            self.neural_net.train(extended_training_set, num_epochs, minibatch_size)
            self.neural_net.generation += 1

            # Battle
            if True:
                arena = Arena(self.game, **training_params.arena_parameters.__dict__)
                alpha_zero_before = AlphaZero(neural_net_before)
                arena_result = arena.battle(alpha_zero_before, self)
                if arena_result.must_improve and not arena_result.is_improvement:
                    self.neural_net = neural_net_before

            self.neural_net.save()

    def self_play(
        self,
        mcts: AlphaZeroMcts,
        temperature_schedule: TemperatureSchedule,
        initial_state: str | list[int] | None = None,
    ) -> list[TrainingData]:
        node: Node[TMove] | None = None
        state: State = self.game.initial_state(initial_state)
        status = state.status()

        self.neural_net.set_to_eval()
        recorded_actions: list[RecordedAction] = []

        while status.is_in_progress:
            state_before = state
            policy = mcts.policy(state_before, node)
            temperature = temperature_schedule.get_temperature(state_before.move_count)
            move = policy.select_move(temperature)
            node = next(node for node in policy.node.children if node.move == move)
            state = state_before.play_move(move)
            recorded_action = RecordedAction(state_before, policy.encoded_policy, move, state)
            recorded_actions.append(recorded_action)
            status = state.status()

        training_set: list[TrainingData] = []
        for ph in recorded_actions:
            encoded_state = ph.state_before.to_feature()
            perspective = ph.state_after.played_by
            outcome = status.outcome(perspective)
            training_data = TrainingData(encoded_state, ph.policy, outcome)
            training_set.append(training_data)

        return training_set

    def self_play_parallel(
        self,
        solver: AlphaZeroMcts,
        num_parallel: int,
        temperature_schedule: TemperatureSchedule,
        initial_state: str | list[int] | None = None,
    ) -> list[TrainingData]:
        initial_states: list[State] = self.game.generate_states(2, num_parallel, initial_state)
        parallel_games = [ParallelGame(s) for s in initial_states if s.status().is_in_progress]
        in_progress_games = list(parallel_games)
        nodes: list[Node[TMove]] | None = None

        self.neural_net.set_to_eval()

        progress_bar = tqdm.tqdm(total=float("inf"), unit=" -- count:", leave=False)
        while in_progress_games:
            states = [pg.latest_state for pg in in_progress_games]
            policies = solver.policies(states, nodes)

            nodes = []
            games_tmp = []
            for policy, pg in zip(policies, in_progress_games):
                policy, pg = cast(Policy, policy), cast(ParallelGame, pg)
                state_before = pg.latest_state
                temperature = temperature_schedule.get_temperature(state_before.move_count)
                move = policy.select_move(temperature)
                node = next(node for node in policy.node.children if node.move == move)
                state = state_before.play_move(move)
                recorded_action = RecordedAction(state_before, policy.encoded_policy, move, state)
                pg.recorded_actions.append(recorded_action)
                if state.status().is_in_progress:
                    games_tmp.append(pg)
                    nodes.append(node)
            in_progress_games = games_tmp

            progress_bar.update(1)
        progress_bar.close()

        training_set: list[TrainingData] = []
        for pg in parallel_games:
            terminal_status = pg.latest_state.status()
            policy_history = pg.recorded_actions
            for ph in policy_history:
                encoded_state = ph.state_before.to_feature()
                perspective = ph.state_after.played_by
                outcome = terminal_status.outcome(perspective)
                training_data = TrainingData(encoded_state, ph.policy, outcome)
                training_set.append(training_data)

        return training_set

    def explore_hyperparameters(self, generation: int, arena_parameters: AZArenaParameters) -> None:
        hps = {
            "minibatch_size": [256, 512, 1024, 2048, 4096],
            "num_epochs": [2, 4, 6, 8, 10, 12],
            "learning_rate": [0.01, 0.001, 0.000_1, 0.000_01],
        }

        writer = SummaryWriter(log_dir=f"runs/{self.game.fullname}_Gen{generation:03d}")
        arena = Arena(self.game, **arena_parameters.__dict__)
        training_set = self.neural_net.load_training_data(generation)

        all_hyper_params = list(product(hps["num_epochs"], hps["minibatch_size"]))
        for num_epochs, minibatch_size in trange(all_hyper_params, desc="Hyperparameter Search"):
            neural_net_after = deepcopy(self.neural_net)

            # Train the network with the current hyperparameters
            neural_net_after.train(training_set, num_epochs, minibatch_size, log_progress=False)
            alpha_zero_after = AlphaZero(neural_net_after)
            arena_result = arena.battle(self, alpha_zero_after, log_progress=False)

            hparams = {"minibatch_size": minibatch_size, "num_epochs": num_epochs}
            metrics = {"win_ratio": arena_result.win_ratio}
            writer.add_hparams(hparams, metrics)

        writer.close()

    def _get_mcts(self, num_mcts_sims: int | AZMctsParameters) -> AlphaZeroMcts:
        if isinstance(num_mcts_sims, int):
            default_params = AZMctsParameters.defaults(self.game.fullname)
            value_by_prop = asdict(default_params)
            value_by_prop["num_mcts_sims"] = num_mcts_sims
        else:
            value_by_prop = asdict(num_mcts_sims)

        return AlphaZeroMcts(self.neural_net, **value_by_prop)

    def _exploit_symmetries(self, training_set: list[TrainingData]) -> list[TrainingData]:
        def all_sets(data: TrainingData) -> list[TrainingData]:
            symmetries = self.game.symmetries(data.encoded_state, data.policy)
            return (TrainingData(state, pol, data.outcome) for state, pol in symmetries)

        return [sym_data for training_data in training_set for sym_data in all_sets(training_data)]

    def _get_random_state(
        self, num_moves: int, probability: float, initial_state: str | list[int] | None = None
    ) -> State:
        state: State = self.game.initial_state(initial_state)
        if np.random.rand() > probability:
            return state

        for _ in range(num_moves):
            status = state.status()
            if status.is_in_progress:
                move = np.random.choice(status.legal_moves)
                state.play_move(move)

        return state


@dataclass(slots=True)
class AlphaZeroSolver(Solver):
    alpha_zero: AlphaZero
    num_mcts_sims: int

    def solve(self, state: State[TMove]) -> TMove:
        return self.alpha_zero.select_move(state, self.num_mcts_sims)


class Arena:
    def __init__(
        self,
        game: Game,
        *,
        num_games: int,
        num_mcts_sims: int,
        required_win_ratio: float,
        temperature_schedule: TemperatureSchedule,
    ) -> None:
        self.game = game
        self.num_games = num_games
        self.num_mcts_sims = num_mcts_sims
        self.required_win_ratio = required_win_ratio
        self.temperature_schedule = temperature_schedule

    def battle(
        self, current: AlphaZero, challenger: AlphaZero, log_progress: bool = True
    ) -> ArenaResult:
        games_to_play = self.num_games // 2

        num_wins1, num_draws1, num_losses1 = self.play_games(current, challenger, games_to_play)
        num_wins2, num_draws2, num_losses2 = self.play_games(challenger, current, games_to_play)

        arena_result = ArenaResult(
            challenger.neural_net.generation,
            num_wins2 + num_losses1,
            num_draws1 + num_draws2,
            num_losses2 + num_wins1,
            self.required_win_ratio,
        )

        if log_progress:
            self._log_result(arena_result)

        return arena_result

    def play_games(
        self, player1: AlphaZero, player2: AlphaZero, games_to_play: int
    ) -> tuple[float, float, float]:
        initial_states: list[State] = self.game.generate_states(2, games_to_play)

        active_states = [state for state in initial_states if state.status().is_in_progress]
        finished_states: list[State] = []

        player, opponent = player1, player2
        while active_states:
            moves = player.select_moves(active_states, self.num_mcts_sims, self.temperature_schedule)
            for i, move in enumerate(moves):
                active_states[i] = active_states[i].play_move(move)

            player, opponent = opponent, player

            statuses = [state.status().is_in_progress for state in active_states]
            finished_states += [s for s, in_progress in zip(active_states, statuses) if not in_progress]
            active_states = [s for s, in_progress in zip(active_states, statuses) if in_progress]

        outcomes = [state.status().outcome(1) for state in finished_states]
        num_wins = len([outcome for outcome in outcomes if int(outcome) == 1])
        num_draws = len([outcome for outcome in outcomes if int(outcome) == 0])
        num_losses = games_to_play - num_wins - num_draws
        return num_wins, num_draws, num_losses

    def _log_result(self, arena_result: ArenaResult) -> None:
        writer = SummaryWriter(log_dir=f"runs/{self.game.fullname}")
        gen = arena_result.generation

        # Logging metrics to TensorBoard
        writer.add_scalar(f"Arena/Win Ratio", arena_result.win_ratio, gen)
        writer.add_scalar(f"Arena/Total Wins", arena_result.num_wins, gen)
        writer.add_scalar(f"Arena/Total Draws", arena_result.num_draws, gen)
        writer.add_scalar(f"Arena/Total Losses", arena_result.num_losses, gen)
        writer.close()


@dataclass
class InitialStateGenerator:
    initial_state: State
    depth: int

    def generate(self):
        pass


@dataclass(slots=True)
class ArenaResult:
    generation: int
    num_wins: float
    num_draws: float
    num_losses: float
    required_ratio_cutoff: float
    must_improve: bool = False

    @property
    def win_ratio(self) -> bool:
        return (self.num_wins + 0.5 * self.num_draws) / self.num_games

    @property
    def num_games(self) -> bool:
        return self.num_wins + self.num_draws + self.num_losses

    @property
    def is_improvement(self) -> bool:
        return self.win_ratio > self.required_ratio_cutoff



from abc import ABC, abstractmethod
from typing import Generic

import numpy as np

from ..solvers.mcts import Node
from ..states import ConnectXState, TicTacToeState, TState
from .plot import heatmap_checker_colors, light_checker_colors


class HtmlBuilder(ABC, Generic[TState]):
    @abstractmethod
    def build_html(self, state: TState) -> str:
        ...

    @abstractmethod
    def build_tiny_html(self, state: TState) -> str:
        ...

    @abstractmethod
    def display(self, state: TState) -> None:
        ...

    def _html(self, grid: np.ndarray, policy: np.ndarray | None, has_numeric_cols: bool) -> str:
        tiny_html = self._tiny_html(
            grid,
            heatmap_values=policy,
            include_headers=True,
            has_numeric_cols=has_numeric_cols,
        )

        html_table = f"""
        <table class="ttt-board" style="margin:auto;text-align:center;float:left">
            <tbody>
                {tiny_html}
            </tbody>
        </table>"""

        html_with_css = (
            """
<!DOCTYPE html>
<html>
    <head>
        <title></title>
        <meta charset="UTF-8">
        <style>
            .ttt-board { border-spacing: 0; border-collapse: collapse; }
            .ttt-board th { padding: .5em; }
            .ttt-board th + th { border-bottom: 1px solid #000; }
            .ttt-board th:first-child,
            .ttt-board td:last-child { border-right: 1px solid #000; }
            .ttt-board tr:last-child td { border-bottom: 1px solid; }
            .ttt-board th:empty { border: none; }
            .ttt-board td { width: 1.5em; height: 1.5em; text-align: center; font-size: 18px; line-height: 0;}
        </style>
    </head>
    <body>"""
            + html_table
            + """
    </body>
</html>
"""
        )

        return html_with_css

    def _tiny_html(
        self,
        grid: np.ndarray,
        heatmap_values: np.ndarray | None = None,
        include_headers: bool = False,
        has_numeric_cols: bool = False,
    ) -> str:
        rows, cols = grid.shape

        pieces = np.vectorize(HtmlBuilder.piece)(grid)

        if heatmap_values is not None:
            colors = heatmap_checker_colors(heatmap_values)
        else:
            colors = light_checker_colors(grid.shape)

        table = ""

        if include_headers:
            table += "<tr>\n"
            table += "<th></th>\n"
            for col in range(cols):
                letter = col + 1 if has_numeric_cols else chr(ord("a") + col)
                table += f"<th><center>{letter}</center></th>\n"
            table += "</tr>\n"

        for row in range(rows):
            table += "<tr>\n"
            if include_headers:
                table += f"<th>{rows - row}</th>\n"
            for col in range(cols):
                mark = pieces[row, col]
                color = colors[row, col]
                table += f'<td height="20" width="20" bgcolor="{color}">{mark}</td>\n'
            table += "</tr>\n"
        return table

    @staticmethod
    def piece(mark) -> str:
        match mark:
            case 1:
                return r'<font POINT-SIZE="12">⭕</font>'
            case 2:
                return r'<font POINT-SIZE="12">✖️</font>'
            case _:
                return ""


class MctsNodeHtmlBuilder:
    def build_tiny_html(self, node: Node, state: TState) -> str:
        _, cols = state.shape

        win_label, ucb_label = MctsNodeHtmlBuilder._get_labels(node)

        tiny = state.html(is_tiny_repr=True)
        html = f"""<<table cellborder="0" cellspacing="0" border="0">
            {tiny}
            <tr>
                <td height="5" colspan="{cols}" align="center"></td>
            </tr>
            <tr>
                <td colspan="{cols}" align="center"><font face="Helvetica" color="#424242" POINT-SIZE="11">{win_label}</font></td>
            </tr>
            <tr>
                <td colspan="{cols}" align="center"><font face="Helvetica" color="#424242" POINT-SIZE="11">{ucb_label}</font></td>
            </tr>
        </table>>"""

        return html

    @staticmethod
    def _get_labels(node: Node) -> tuple[str, str]:
        win_label = f"N={node.visit_count}, W={node.wins}"
        ucb_label = f"UCB={node.ucb():.4}" if node.parent else "root"
        return win_label, ucb_label


class TicTacToeHtmlBuilder(HtmlBuilder[TicTacToeState]):
    def display(self, state: TicTacToeState) -> None:
        from IPython.display import HTML

        html_string = self.build_html(state)
        return HTML(html_string)

    def build_tiny_html(self, state: TicTacToeState) -> str:
        grid = state.to_grid()
        return self._tiny_html(grid)

    def build_html(self, state: TicTacToeState, policy: np.ndarray | None = None) -> str:
        grid = state.to_grid()
        return self._html(grid, policy=policy, has_numeric_cols=False)


class ConnectXHtmlBuilder(HtmlBuilder[ConnectXState]):
    def display(self, state: ConnectXState) -> None:
        from IPython.display import HTML

        html_string = self.build_html(state)
        return HTML(html_string)

    def build_tiny_html(self, state: ConnectXState) -> str:
        grid = state.to_grid()
        return self._tiny_html(grid)

    def build_html(self, state: ConnectXState, policy: np.ndarray | None = None) -> str:
        grid = state.to_grid()
        return self._html(grid, policy=policy, has_numeric_cols=True)



from copy import copy
from dataclasses import dataclass
from enum import Enum
from typing import Generator

import numpy as np
from colorama import Fore, Style

UNKNOWN = 0
EXCLUDE = 1
INCLUDE = 2


class VectorType(Enum):
    ROW = 1
    COLUMN = 2


@dataclass
class KnapsackNumberLine:
    def __init__(self, numbers: np.ndarray, mask: np.ndarray) -> None:
        self.numbers: np.ndarray = numbers
        self.mask: np.ndarray = mask

        self._number_line, self._accessibles = self._build_accessible_numbers()

    def can_reach(self, target: int) -> bool:
        result = self._accessibles[self._number_line == target]
        return len(result) == 1 and result[0]

    def _build_accessible_numbers(self) -> tuple[np.ndarray, np.ndarray]:
        lower_bound, upper_bound = self.get_bounds()
        number_line = np.arange(lower_bound, upper_bound + 1)
        accessibles = number_line == 0

        for num, m in zip(self.numbers, self.mask):
            if m == UNKNOWN:
                shifted = KnapsackNumberLine.shift(accessibles, num)
                accessibles |= shifted

        include_shift = np.sum(self.numbers[self.mask == INCLUDE])
        accessibles = KnapsackNumberLine.shift(accessibles, include_shift)

        return (number_line, accessibles)

    def get_bounds(self) -> tuple[int, int]:
        lower_bound = np.sum(self.numbers[self.numbers < 0])
        upper_bound = np.sum(self.numbers[self.numbers > 0])
        return lower_bound, upper_bound

    @staticmethod
    def shift(x: np.ndarray, n: int) -> np.ndarray:
        if n == 0:
            return x[:]

        if n > 0:
            return np.pad(x, (n, 0))[:-n]

        return np.pad(x, (0, -n))[-n:]


@dataclass
class BoardVector:
    numbers: np.ndarray
    mask: np.ndarray
    target: float
    vector_type: VectorType
    index: int

    def unknowns(self) -> Generator[int, None, None]:
        for i, m in enumerate(self.mask):
            if m == UNKNOWN:
                yield i

    def sum(self) -> int:
        included = self.mask != EXCLUDE
        return np.sum(self.numbers[included])

    def full_index(self, index: int) -> tuple[int, int]:
        if self.vector_type == VectorType.COLUMN:
            return index, self.index
        if self.vector_type == VectorType.ROW:
            return self.index, index
        raise ValueError(f"vector {self.vector_type} type unknown")

    def without(self, index: int) -> KnapsackNumberLine:
        mask = self.mask.copy()
        mask[index] = EXCLUDE
        return KnapsackNumberLine(self.numbers, mask)

    def include(self, index: int) -> None:
        self.mask[index] = INCLUDE

    def exclude(self, index: int) -> None:
        self.mask[index] = EXCLUDE


@dataclass
class Board:
    grid: np.ndarray
    mask: np.ndarray
    row_sums: np.ndarray
    col_sums: np.ndarray

    def unsolved_vectors(self) -> Generator[BoardVector, None, None]:
        yield from self.unsolved_rows()
        yield from self.unsolved_cols()

    def unsolved_rows(self) -> Generator[BoardVector, None, None]:
        for i in range(self.grid.shape[0]):
            row_mask = self.mask[i, :]
            if np.any(row_mask == UNKNOWN):
                grid, mask = self.grid[i, :].view(), self.mask[i, :].view()
                yield BoardVector(grid, mask, self.row_sums[i], VectorType.ROW, i)

    def unsolved_cols(self) -> Generator[BoardVector, None, None]:
        for j in range(self.grid.shape[1]):
            col_mask = self.mask[:, j]
            if np.any(col_mask == UNKNOWN):
                grid, mask = self.grid[:, j], self.mask[:, j]
                yield BoardVector(grid, mask, self.col_sums[j], VectorType.COLUMN, j)

    def is_solved(self) -> bool:
        included = self.mask != EXCLUDE
        col_sums = np.sum(self.grid * included, axis=0)
        row_sums = np.sum(self.grid * included, axis=1)

        return np.alltrue(col_sums == self.col_sums) and np.alltrue(row_sums == self.row_sums)

    def __copy__(self):
        return Board(self.grid.copy(), self.mask.copy(), self.row_sums.copy(), self.col_sums.copy())

    @classmethod
    def create(cls, size: int) -> "Board":
        rng = np.random.RandomState()

        if size > 8:
            value_range = np.concatenate((np.arange(-20, 0), np.arange(1, 21)))
            grid = np.random.choice(value_range, size=(size, size))
        else:
            grid = rng.randint(low=1, high=10, size=(size, size))

        true_mask = np.random.randint(low=1, high=3, size=(size, size))
        included = grid * (true_mask == INCLUDE)
        row_sums = np.sum(included, axis=1)
        col_sums = np.sum(included, axis=0)

        return cls(grid, grid * 0, row_sums, col_sums)


class BoardPrinter:
    def print(self, board: Board) -> None:
        rows, cols = board.grid.shape

        # Print header row
        col_nums = "    " + " ".join([f" {i+1:2}" for i in range(cols)])
        row_divider = "   |" + "-" * (cols * 4) + "|"
        print(col_nums)
        print(row_divider)

        # Print each row of the matrix
        for i in range(rows):
            row_items = []
            for j in range(cols):
                item = f"{board.grid[i][j]:3}"
                if board.mask[i][j] == 1:
                    item = f"{Fore.RED}{item}{Style.RESET_ALL}"
                elif board.mask[i][j] == 2:
                    item = f"{Fore.GREEN}{item}{Style.RESET_ALL}"
                row_items.append(item)

            row_string = " ".join(row_items)
            print(f"{i+1:2} |{row_string} |{board.row_sums[i]:3}")

        # Print the footer for the matrix
        col_sums = [f"{board.col_sums[j]:3}" for j in range(cols)]
        col_string = " ".join(col_sums)
        print(row_divider)
        print(f"   |{col_string} |")


class SumpleteSolver:
    def solve(self, board: Board) -> bool:
        is_updating = True
        while is_updating:
            is_updating = False
            for vector in board.unsolved_vectors():
                for index in vector.unknowns():
                    number_line = vector.without(index)
                    if not number_line.can_reach(vector.target):
                        vector.include(index)
                        is_updating = True
                    elif not number_line.can_reach(vector.target - vector.numbers[index]):
                        vector.exclude(index)
                        is_updating = True

                # If the target cannot be reached...
                if np.all(vector.mask != UNKNOWN) and vector.sum() != vector.target:
                    return False

        if board.is_solved():
            return True

        # Now we recurse...
        def backtrack(try_include: bool) -> bool:
            speculative_board = copy(board)
            vector = next(speculative_board.unsolved_vectors())
            index = next(vector.unknowns())

            if try_include:
                vector.include(index)
            else:
                vector.exclude(index)

            if self.solve(speculative_board):
                board.mask = speculative_board.mask
                return True

            return False

        return backtrack(True) or backtrack(False)


if __name__ == "__main__":
    n = 8
    board = Board.create(n)

    solver = SumpleteSolver()
    if solver.solve(board):
        printer = BoardPrinter()
        printer.print(board)
    else:
        print("failed to solve")



from dataclasses import dataclass
from typing import Generator, Self

import numpy as np

from ..bitboard_utils import BitboardUtil
from .state import State, Status, TemperatureSchedule


@dataclass(slots=True)
class ConnectXState(State[int]):
    bitboard_util: BitboardUtil
    mask: int
    position: int
    num_moves: int

    @property
    def rows(self) -> int:
        return self.bitboard_util.rows - 1

    @property
    def cols(self) -> int:
        return self.bitboard_util.cols

    @property
    def shape(self) -> tuple[int, int]:
        return self.rows, self.cols

    @property
    def num_slots(self) -> int:
        return self.rows * self.cols

    @property
    def is_full(self) -> bool:
        return self.num_moves == self.num_slots

    @property
    def played_by(self) -> int:
        is_odd_num_moves = self.num_moves & 1
        return 2 - is_odd_num_moves

    @property
    def move_count(self) -> int:
        return self.num_moves

    def status(self) -> Status[int]:
        is_won = self.is_won()
        is_ended = is_won or self.is_full

        is_in_progress = not is_ended
        value = 1.0 if is_won else 0.0
        legal_moves = [] if is_ended else list(self._possible_moves_unchecked())
        return Status(is_in_progress, self.played_by, value, legal_moves)

    def can_play_move(self, col: int) -> bool:
        offset = (col + 1) * self.bitboard_util.rows - 2
        top_col_bit = 1 << offset
        return (self.mask & top_col_bit) == 0

    def play_move(self, col: int) -> Self:
        state = self.__copy__()
        state.set_move(col)
        return state

    def set_move(self, col: int) -> None:
        offset = self.bitboard_util.rows * col
        col_bit = 1 << offset
        bitmove = self.mask + col_bit
        self.set_bitmove(bitmove)

    def set_bitmove(self, bitmove: int) -> None:
        self.position ^= self.mask
        self.mask |= bitmove
        self.num_moves += 1

    def policy_loc(self, move: int) -> int:
        return move

    def key(self) -> int:
        return (self.mask + self.bitboard_util.BOTTOM_ROW) | self.position

    def outcome(self, perspective: int, indicator: str = "win-loss") -> float:
        score = self._outcome(perspective)
        if indicator == "win-loss":
            return (1 + np.sign(score)) / 2

        return score

    def _outcome(self, perspective: int) -> int:
        if self.is_full and not self.is_won():
            return 0

        score = (self.num_slots - self.num_moves + 2) // 2
        is_odd_num_moves = self.num_moves & 1
        is_odd_perspective = perspective & 1

        return score if is_odd_num_moves == is_odd_perspective else -score

    def is_won(self) -> bool:
        rows = self.rows + 1
        directions = (1, rows - 1, rows, rows + 1)
        bitboard = self.position ^ self.mask
        for dir in directions:
            bitmask = bitboard & (bitboard >> dir)
            if bitmask & (bitmask >> 2 * dir):
                return True

        return False

    def legal_bitmoves(self) -> Generator[int, None, None]:
        if self.is_won() or self.is_full:
            return range(0)

        return self._possible_bitmoves_unchecked()

    def win_mask(self) -> int:
        H1 = self.bitboard_util.rows
        posn = self.position

        # Vertical win
        wm = (posn << 1) & (posn << 2) & (posn << 3)

        # Horizontals (_XXXO and OXXX_)
        wm |= (posn << H1) & (posn << 2 * H1) & (posn << 3 * H1)
        wm |= (posn >> H1) & (posn >> 2 * H1) & (posn >> 3 * H1)

        # Horizontals (OXX_XO and OX_XXO)
        wm |= (posn << H1) & (posn << 2 * H1) & (posn >> H1)
        wm |= (posn >> H1) & (posn >> 2 * H1) & (posn << H1)

        # Diagonals _/_
        wm |= (posn << (H1 + 1)) & (posn << 2 * (H1 + 1)) & (posn << 3 * (H1 + 1))
        wm |= (posn << (H1 + 1)) & (posn << 2 * (H1 + 1)) & (posn >> (H1 + 1))
        wm |= (posn << (H1 + 1)) & (posn >> (H1 + 1)) & (posn >> 2 * (H1 + 1))
        wm |= (posn >> (H1 + 1)) & (posn >> 2 * (H1 + 1)) & (posn >> 3 * (H1 + 1))

        # Diagonals _\_
        wm |= (posn >> (H1 - 1)) & (posn >> 2 * (H1 - 1)) & (posn >> 3 * (H1 - 1))
        wm |= (posn >> (H1 - 1)) & (posn >> 2 * (H1 - 1)) & (posn << (H1 - 1))
        wm |= (posn >> (H1 - 1)) & (posn << (H1 - 1)) & (posn << 2 * (H1 - 1))
        wm |= (posn << (H1 - 1)) & (posn << 2 * (H1 - 1)) & (posn << 3 * (H1 - 1))

        return wm & (self.bitboard_util.BOARD_MASK ^ self.mask)

    def to_grid(self) -> np.ndarray:
        rows, cols = self.shape
        sequence = np.arange((rows + 1) * cols, dtype=object)
        indices = np.flipud(sequence.reshape((cols, rows + 1)).T)
        powers = 2 ** indices[1:]

        posn = self.position ^ self.mask if self.num_moves & 1 else self.position
        player_1 = posn
        player_2 = posn ^ self.mask
        r = np.sign(player_1 & powers).astype(np.int8)
        g = np.sign(player_2 & powers).astype(np.int8)
        return np.asarray(r + 2 * g)

    def to_feature(self) -> np.ndarray:
        rows, cols = self.shape
        sequence = np.arange((rows + 1) * cols, dtype=object)
        indices = np.flipud(sequence.reshape((cols, rows + 1)).T)
        powers = 2 ** indices[1:]

        player_1 = self.position
        player_2 = self.position ^ self.mask
        r = np.sign(player_1 & powers).astype(np.float32)
        g = np.sign(player_2 & powers).astype(np.float32)
        b = 1 - r - g
        return np.stack((r, g, b))

    def get_input_move(self) -> int:
        while True:
            response = input("Your turn to move. Please enter an integer: ")
            try:
                move = int(response) - 1
                if move in self.status().legal_moves:
                    return move
            except ValueError:
                pass
            print("Move was invalid. Please try again.")

    def html(self, policy: np.ndarray | None = None, is_tiny_repr: bool = False) -> str:
        from ..views.html import ConnectXHtmlBuilder

        html_printer = ConnectXHtmlBuilder()

        if is_tiny_repr:
            return html_printer.build_tiny_html(self)

        return html_printer.build_html(self, policy)

    def plot(self):
        from ..views.plot import plot_state

        plot_state(self.to_feature(), figsize=(4, 4))

    def _possible_moves_unchecked(self) -> Generator[int, None, None]:
        possible_moves_mask = self._possible_bitmoves_mask()
        move_order = self.bitboard_util.move_order()

        for col in move_order:
            col_mask = self.bitboard_util.get_col_mask(col)
            possible_move = possible_moves_mask & col_mask
            if possible_move:
                yield col

    def _possible_bitmoves_unchecked(self) -> Generator[int, None, None]:
        possible_moves_mask = self._possible_bitmoves_mask()
        move_order = self.bitboard_util.move_order()

        for col in move_order:
            col_mask = self.bitboard_util.get_col_mask(col)
            possible_move = possible_moves_mask & col_mask
            if possible_move:
                yield possible_move

    def _possible_bitmoves_mask(self) -> int:
        return (self.mask + self.bitboard_util.BOTTOM_ROW) & self.bitboard_util.BOARD_MASK

    def _repr_html_(self) -> str:
        return self.html()

    def __copy__(self) -> "ConnectXState":
        return ConnectXState(self.bitboard_util, self.mask, self.position, self.num_moves)

    @classmethod
    def from_grid(cls, grid: np.ndarray) -> Self:
        rows, cols = grid.shape
        padded_grid = np.vstack((np.zeros(cols), grid))

        indices = np.flipud(np.arange((rows + 1) * cols).reshape((cols, rows + 1)).transpose())
        binary_vals = 2 ** indices.astype(np.int64)

        mask = (padded_grid > 0).astype(np.int64)
        num_moves = np.sum(mask)
        mark = 1 + num_moves % 2

        posn = (padded_grid == mark).astype(np.int64)
        mask_utils = BitboardUtil(rows + 1, cols)
        board = cls(mask_utils, 0, 0, 0)
        mask_val = np.sum(mask * binary_vals)
        posn_val = np.sum(posn * binary_vals)
        board.mask = mask_val
        board.position = posn_val
        board.num_moves = int(np.sum(mask))
        return board

    @classmethod
    def create(cls, rows: int, cols: int, moves: str | list[int] | None = None) -> Self:
        util = BitboardUtil(rows + 1, cols)
        state = cls(util, 0, 0, 0)
        moves = moves or []

        if isinstance(moves, str):
            move_array = moves.replace(" ", "").split(",")
            moves = [int(move) for move in move_array]

        for move in moves:
            state.set_move(move - 1)
        return state



from __future__ import annotations

import asyncio
import logging
import tkinter as tk
from asyncio import Event
from importlib import resources
from tkinter import ttk

import chess
from chess import Piece, Square
from PIL import Image, ImageTk

from caz.alpha_zero import AlphaZero, PytorchNeuralNetwork
from caz.games import Chess
from caz.states import ChessState
from caz.states.state import TemperatureSchedule

ROWS, COLS = 8, 8
CELL_SIZE = 80
REFRESH_RATE = 0.025

BLACK_COLOR = "#EDF9EB"
WHITE_COLOR = "#346B51"

DARK_COLOR = "#1F1F1F"
TEAL_COLOR = "#78C5B0"
BLUE_COLOR = "#77BEFC"
MATE_COLOR = "#880015"
DRAW_COLOR = "#FFA628"
LAST_COLOR_L = "#F8F77E"
LAST_COLOR_D = "#BECC52"
HIGH_COLOR_L = "#8DD1BA"
HIGH_COLOR_D = "#349688"
HIGH_COLOR = "#B5E61D"


class ChessScreenModel:
    def __init__(self, alpha_zero: AlphaZero, colors: list[chess.Color]) -> None:
        self.alpha_zero = alpha_zero
        self.colors = colors
        self.root_node = None
        self.event: Event | None = None
        self.is_pondering: bool = False

    def is_alphas_turn(self, state: ChessState) -> bool:
        return state.board.turn in self.colors

    def reset(self, colors: list[chess.Color]) -> None:
        self.root_node = None
        self.colors = colors

    async def ponder(self, state: ChessState) -> None:
        if not self.colors:
            self.is_pondering = False
            return

        PONDER_SIMS = 50_000
        self.is_pondering = True
        await asyncio.sleep(REFRESH_RATE)
        self.event = Event()
        policy = await self.alpha_zero.policy_async(state, PONDER_SIMS, self.root_node, self.event)
        self.root_node = policy.node
        self.is_pondering = False

    def stop_pondering(self) -> None:
        if self.event:
            self.event.set()
            self.event = None

    async def wait_for_pondering_to_stop(self) -> None:
        while self.is_pondering:
            await asyncio.sleep(REFRESH_RATE)

    async def get_move(self, state: ChessState, num_sims: int) -> tuple[chess.Move, float]:
        node = None
        if len(self.colors) == 2:
            node = self.root_node
        elif self.root_node and state.board.move_stack:
            node = next(n for n in self.root_node.children if n.move == state.board.move_stack[-1])

        temperature_schedule = TemperatureSchedule(2, 1)
        await asyncio.sleep(REFRESH_RATE)
        policy = await self.alpha_zero.policy_async(state, num_sims, node)
        temperature = temperature_schedule.get_temperature(state.move_count)
        move = policy.select_move(temperature)
        self.root_node = next(node for node in policy.node.children if node.move == move)
        q_value = self.root_node.q_value
        return move, q_value if self.root_node.played_by == 1 else 1 - q_value


class ChessScreenController:
    def __init__(self, model: ChessScreenModel, view: ChessScreen):
        self.model = model
        self.view = view

        self.state = ChessState.create()

        self.selected_square: Square | None = None

        self.view.on_square_click = self.on_square_click
        self.view.draw_board(self.state)

    @property
    def num_sims(self) -> int:
        return int(self.view.sims_combobox.get())

    async def reset(self, human_colors: list[chess.Color]) -> None:
        colors_complement = []
        if chess.WHITE not in human_colors:
            colors_complement.append(chess.WHITE)
        if chess.BLACK not in human_colors:
            colors_complement.append(chess.BLACK)

        self.state = ChessState.create()
        self.model.reset(colors_complement)
        self.view.set_players("Human", human_colors)
        self.view.set_players("CAZ", colors_complement)
        self.view.draw_board(self.state)
        self.view.update_eval_bar(0.5)
        if self.state.status().is_in_progress and self.model.is_alphas_turn(self.state):
            move, q_value = await self.model.get_move(self.state, self.num_sims)
            await self.move_piece(move, q_value)

    async def on_square_click(self, square: Square) -> None:
        if self.model.is_alphas_turn(self.state) or not self.state.status().is_in_progress:
            return

        piece = self.state.board.piece_at(square)

        if piece and piece.color == self.state.board.turn and square != self.selected_square:
            self.selected_square = square
            self.view.highlight_click(square)

        elif self.selected_square is not None:
            move = chess.Move(self.selected_square, square)
            await self.move_piece(move)

        else:
            self.view.draw_board(self.state)

    async def move_piece(self, move: chess.Move, q_value: float | None = None):
        from_square = move.from_square
        to_square = move.to_square
        piece = self.state.board.piece_at(from_square)
        self.selected_square = None

        if move.promotion is None and piece.piece_type == chess.PAWN:
            promotion_rank = 7 if self.state.board.turn else 0
            if chess.square_rank(to_square) == promotion_rank:
                move.promotion = self.view.promotion_option.get()

        # Check if the move is legal
        if piece.color == self.state.board.turn and move in self.state.board.legal_moves:
            self.model.stop_pondering()
            await self.model.wait_for_pondering_to_stop()
            self.state.board.push(move)
            self.view.draw_board(self.state)
            self.view.update_eval_bar(q_value)

            await asyncio.sleep(0)
            csv = self.state.csv()
            logging.info(csv)

            if not self.state.status().is_in_progress:
                await asyncio.sleep(0)
                pgn = self.state.pgn()
                logging.info("\n" + pgn)
                return

            if self.model.is_alphas_turn(self.state):
                move, q_value = await self.model.get_move(self.state, self.num_sims)
                await self.move_piece(move, q_value)
            else:
                await self.model.ponder(self.state)
        else:
            self.view.draw_board(self.state)


class ChessScreen(tk.Frame):
    def update_eval_bar(self, q_value: float | None) -> None:
        if q_value is None:
            return

        canvas_width = 20
        canvas_height = 8 * CELL_SIZE
        self.eval_bar.delete("all")

        bar_height = int(canvas_height)  # Height of each half bar
        bar_width = 20  # Width of the bar
        bar_position = int((canvas_width - bar_width) / 2)  # X-coordinate for the bar

        # Calculate the height of the black and white halves based on the given value
        white_height = int(bar_height * q_value)
        black_height = bar_height - white_height

        # Render the black half bar
        x1, y1, x2, y2 = bar_position, 0, bar_position + bar_width, black_height
        self.eval_bar.create_rectangle(x1, y1, x2, y2, fill="black")

        # Render the white half bar
        x1, y1, x2, y2 = bar_position, black_height, bar_position + bar_width, canvas_height
        self.eval_bar.create_rectangle(x1, y1, x2, y2, fill="white")

    def build_eval_bar(self, root) -> tk.Canvas:
        canvas_width = 20
        canvas_height = 8 * CELL_SIZE

        canvas = tk.Canvas(root, width=canvas_width, height=canvas_height, highlightthickness=0)
        canvas.pack(expand=True)
        return canvas

    def toggle_chess_bar(self) -> None:
        if self.eval_var.get():
            self.eval_bar.pack()
        else:
            self.eval_bar.pack_forget()

    def __init__(self, parent, new_game_callback):
        super().__init__(parent, bg=DARK_COLOR)

        # Initialize variables
        self.image_map = ChessScreen._get_piece_image_map()

        # Eval Bar
        eval_frame = tk.Frame(self, bg=DARK_COLOR)
        eval_frame.pack(side=tk.LEFT, padx=15, anchor=tk.W, expand=True)
        self.eval_bar = self.build_eval_bar(eval_frame)
        self.update_eval_bar(0.5)

        # LHS Canvas
        self.canvas = tk.Canvas(self, width=8 * CELL_SIZE, height=8 * CELL_SIZE, highlightthickness=0)
        self.canvas.pack(side=tk.LEFT, padx=0, pady=25, expand=False)
        self.canvas.bind("<Button-1>", self._canvas_callback)

        # RHS Frame
        rhs_frame = tk.Frame(self, bg=DARK_COLOR)
        rhs_frame.pack(side=tk.LEFT, padx=20, anchor=tk.W, expand=True)

        # Add labels to display game information
        self.flip_board = tk.BooleanVar(value=False)
        self.eval_var = tk.BooleanVar(value=True)
        self.white_label, self.black_label = self._add_status_labels(rhs_frame)

        # Create radio buttons for promotion options
        self.promotion_option = tk.IntVar(value=chess.QUEEN)
        self._add_promotion_options(rhs_frame)

        # Add new game button
        self._add_catchemalphazero_logo(rhs_frame)
        self._add_new_game_button(rhs_frame, new_game_callback)

    async def on_square_click(self, square: Square) -> None:
        ...

    def draw_board(self, state: ChessState):
        self.canvas.delete("all")

        # Big white square
        self.canvas.create_rectangle(
            0,
            0,
            8 * CELL_SIZE,
            8 * CELL_SIZE,
            fill=WHITE_COLOR,
            outline="",
            tags="cell_rect",
        )

        # Overlay the black squares
        for square in range(ROWS * COLS):
            file, rank = chess.square_file(square), chess.square_rank(square)
            if (file + rank) % 2 == 1:
                continue
            x1, y1 = file * CELL_SIZE, rank * CELL_SIZE
            x2, y2 = x1 + CELL_SIZE, y1 + CELL_SIZE
            self.canvas.create_rectangle(x1, y1, x2, y2, fill=BLACK_COLOR, outline="", tags="cell_rect")

        for rank in range(ROWS):
            for file in range(COLS):
                square = chess.square(file, 7 - rank)
                square = self._flip_square(square)

                piece = state.board.piece_at(square)
                if piece:
                    self.canvas.create_image(
                        file * CELL_SIZE + CELL_SIZE // 2,
                        rank * CELL_SIZE + CELL_SIZE // 2,
                        image=self.image_map[piece],
                    )

        if state.board.move_stack:
            last_move = state.board.move_stack[-1]
            self.highlight_last_move(last_move)

        status = state.status()
        if not status.is_in_progress:
            self.highlight_mate(state) if status.value > 0 else self.highlight_draw(state)

    def highlight_click(self, square: Square) -> None:
        file, rank = chess.square_file(square), chess.square_rank(square)
        color = HIGH_COLOR_L if (file + rank) & 1 else HIGH_COLOR_D
        self.highlight_cell(file, rank, color, True)

    def highlight_last_move(self, move: chess.Move) -> None:
        from_square = move.from_square
        file, rank = chess.square_file(from_square), chess.square_rank(from_square)
        color = HIGH_COLOR_L if (file + rank) & 1 else HIGH_COLOR_D
        self.highlight_cell(file, rank, color, True, "move_highlight")

        to_square = move.to_square
        file, rank = chess.square_file(to_square), chess.square_rank(to_square)
        color = HIGH_COLOR_L if (file + rank) & 1 else HIGH_COLOR_D
        self.highlight_cell(file, rank, color, False, "move_highlight")

    def highlight_draw(self, state: ChessState) -> None:
        king_square = state.board.king(state.board.turn)
        file, rank = chess.square_file(king_square), chess.square_rank(king_square)
        self.highlight_cell(file, rank, DRAW_COLOR, True)

        king_square = state.board.king(not state.board.turn)
        file, rank = chess.square_file(king_square), chess.square_rank(king_square)
        self.highlight_cell(file, rank, DRAW_COLOR, False)

    def highlight_mate(self, state: ChessState) -> None:
        king_square = state.board.king(state.board.turn)
        file, rank = chess.square_file(king_square), chess.square_rank(king_square)
        self.highlight_cell(file, rank, MATE_COLOR, True)

    def highlight_cell(
        self, file: int, rank: int, color: str, delete_canvas: bool, tag: str = "highlight"
    ):
        file, rank = self._flip_coords(file, rank)

        x1 = file * CELL_SIZE
        y1 = (7 - rank) * CELL_SIZE
        x2 = x1 + CELL_SIZE
        y2 = y1 + CELL_SIZE

        if delete_canvas:
            self.canvas.delete(tag)
        highlighted_cell = self.canvas.create_rectangle(
            x1 + 2,
            y1 + 2,
            x2 - 2,
            y2 - 2,
            fill=color,
            outline=color,
            width=4,
            tag=tag,
        )
        self.canvas.tag_raise(highlighted_cell, "cell_rect")

    def set_players(self, label: str, colors: list[chess.Color]) -> None:
        if chess.WHITE in colors:
            self.white_label.config(text=f"White: {label}")
        if chess.BLACK in colors:
            self.black_label.config(text=f"Black: {label}")

    def _add_status_labels(self, frame) -> tk.Label:
        players_label = tk.Label(
            frame,
            text="Status",
            font=("Cascadia Mono", 12),
            fg=TEAL_COLOR,
            bg=DARK_COLOR,
        )
        players_label.pack(side=tk.TOP, pady=0, anchor=tk.W)

        white_player_label = tk.Label(
            frame,
            text=f"White: <<PLACEHOLDER>>",
            font=("Cascadia Mono", 11),
            fg=BLUE_COLOR,
            bg=DARK_COLOR,
        )
        white_player_label.pack(side=tk.TOP, pady=(0, 0), anchor=tk.W)

        black_player_label = tk.Label(
            frame,
            text=f"Black: <<PLACEHOLDER>>",
            font=("Cascadia Mono", 11),
            fg=BLUE_COLOR,
            bg=DARK_COLOR,
        )
        black_player_label.pack(side=tk.TOP, pady=0, anchor=tk.W)

        checkbox = tk.Checkbutton(frame, text="Flip Board", variable=self.flip_board)
        checkbox.configure(background=DARK_COLOR, foreground=BLUE_COLOR, font=("Cascadia Mono", 11))
        checkbox.pack(side=tk.TOP, pady=0, anchor=tk.W)

        # Check box for eval bar
        eval_checkbox = tk.Checkbutton(
            frame, text="Show Eval Bar", variable=self.eval_var, command=self.toggle_chess_bar
        )
        eval_checkbox.configure(background=DARK_COLOR, foreground=BLUE_COLOR, font=("Cascadia Mono", 11))
        eval_checkbox.pack(side=tk.TOP, pady=0, anchor=tk.W)

        return white_player_label, black_player_label

    def _add_promotion_options(self, frame) -> None:
        piece_map = {
            "Queen": chess.QUEEN,
            "Knight": chess.KNIGHT,
            "Rook": chess.ROOK,
            "Bishop": chess.BISHOP,
        }

        self.status_label = tk.Label(
            frame,
            text="Promotion Piece",
            font=("Cascadia Mono", 12),
            fg=TEAL_COLOR,
            bg=DARK_COLOR,
        )
        self.status_label.pack(side=tk.TOP, anchor=tk.W, pady=(20, 0))

        for text, value in piece_map.items():
            radio = tk.Radiobutton(
                frame,
                text=text,
                variable=self.promotion_option,
                value=value,
                font=("Cascadia Mono", 11),
                bg=DARK_COLOR,
                fg=BLUE_COLOR,
            )
            radio.pack(anchor=tk.W, side=tk.TOP)

        self.simluation_label = tk.Label(
            frame,
            text="MCTS Num Sims",
            font=("Cascadia Mono", 12),
            fg=TEAL_COLOR,
            bg=DARK_COLOR,
        )
        self.simluation_label.pack(side=tk.TOP, anchor=tk.W, pady=(20, 0))

        # Create a Combobox
        values = (10, 50, 100, 200, 500, 1000, 2000, 5000, 10000)
        self.sims_combobox = ttk.Combobox(frame, values=values, state="readonly")
        self.sims_combobox.current(4)
        self.sims_combobox.pack(side=tk.TOP, anchor=tk.W, pady=(10, 0))

    def _add_catchemalphazero_logo(self, frame) -> None:
        with resources.path("caz.images", "caz_flat_logo.png") as image_path:
            image = Image.open(image_path)

        image = image.resize((180, 176))
        photo = ImageTk.PhotoImage(image)
        image_label = tk.Label(frame, image=photo, bg=DARK_COLOR)
        image_label.image = photo  # Store a reference to avoid garbage collection
        image_label.pack(side=tk.TOP, pady=(20, 20), anchor=tk.W)

    def _add_new_game_button(self, frame, new_game_callback) -> None:
        # Create a custom style for the buttons
        style = ttk.Style()
        style.theme_use("clam")  # put the theme name here, that you want to use
        style.configure(
            "W.TButton",
            foreground=DARK_COLOR,
            background=BLUE_COLOR,
            font=("Cascadia Mono", 12),
            relief=tk.FLAT,
            anchor=tk.CENTER,
        )

        # Create the "Go to Title Screen" button
        new_game_button = ttk.Button(
            frame,
            text="New Game",
            command=new_game_callback,
            style="W.TButton",
            compound=tk.CENTER,
        )
        new_game_button.pack(side=tk.BOTTOM, anchor=tk.S, pady=(0, 10), padx=(0, 0))

    def _canvas_callback(self, event) -> None:
        file = event.x // CELL_SIZE
        rank = 7 - (event.y // CELL_SIZE)
        square = chess.square(file, rank)
        square = self._flip_square(square)
        asyncio.create_task(self.on_square_click(square))

    def _flip_square(self, square: Square) -> Square:
        return 63 - square if self.flip_board.get() else square

    def _flip_coords(self, file: int, rank: int) -> tuple[int, int]:
        if self.flip_board.get():
            return 7 - file, 7 - rank
        return file, rank

    @staticmethod
    def _get_piece_image_map() -> dict[Piece, tk.PhotoImage]:
        with resources.files("caz.images") as dir:
            piece_images = {
                Piece(chess.PAWN, chess.WHITE): Image.open(dir / "white-pawn.png"),
                Piece(chess.ROOK, chess.WHITE): Image.open(dir / "white-rook.png"),
                Piece(chess.KNIGHT, chess.WHITE): Image.open(dir / "white-knight.png"),
                Piece(chess.BISHOP, chess.WHITE): Image.open(dir / "white-bishop.png"),
                Piece(chess.QUEEN, chess.WHITE): Image.open(dir / "white-queen.png"),
                Piece(chess.KING, chess.WHITE): Image.open(dir / "white-king.png"),
                Piece(chess.PAWN, chess.BLACK): Image.open(dir / "black-pawn.png"),
                Piece(chess.ROOK, chess.BLACK): Image.open(dir / "black-rook.png"),
                Piece(chess.KNIGHT, chess.BLACK): Image.open(dir / "black-knight.png"),
                Piece(chess.BISHOP, chess.BLACK): Image.open(dir / "black-bishop.png"),
                Piece(chess.QUEEN, chess.BLACK): Image.open(dir / "black-queen.png"),
                Piece(chess.KING, chess.BLACK): Image.open(dir / "black-king.png"),
            }

        resized_images = {}
        SCALAR = 1.2
        for piece, image in piece_images.items():
            new_width = int(image.width * SCALAR)
            new_height = int(image.height * SCALAR)
            resized_image = image.resize((new_width, new_height), Image.LANCZOS)
            photo_image = ImageTk.PhotoImage(resized_image)
            resized_images[piece] = photo_image

        return resized_images


class Application(tk.Tk):
    def __init__(self):
        super().__init__()

        self.title("CatchemAlphaZero")
        self.resizable(False, False)
        self.iconbitmap("images/tiny_chess.ico")
        self.is_open = True

        game = Chess()
        network = PytorchNeuralNetwork.create(game, ".")
        alpha_zero = AlphaZero(network)
        model = ChessScreenModel(alpha_zero, [chess.BLACK])

        self.game_screen = ChessScreen(self, lambda: print("Single Window!"))
        self.controller = ChessScreenController(model, self.game_screen)
        self.game_screen.pack(fill=tk.BOTH, expand=True)

        self.protocol("WM_DELETE_WINDOW", self.raise_exit_flag)

    async def show(self):
        while self.is_open:
            self.update()
            await asyncio.sleep(0.025)

    def raise_exit_flag(self):
        self.is_open = False


if __name__ == "__main__":
    app = Application()
    asyncio.run(app.show())



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Generic

import numpy as np

from ..kaggle_types import Configuration, Observation
from ..solvers import Solver
from ..states import ChessState, ConnectXState, TicTacToeState, TState
from ..states.chess_enums import ActionPlanes, ObsPlanes
from ..views import View
from ..views.console import ChessTicTacToeView, ConsoleConnectXView, ConsoleTicTacToeView
from .game_type import GameType


@dataclass(frozen=True, slots=True)
class GameParameters:
    shape: tuple[int, int]
    observation_shape: tuple[int, int, int]
    action_size: int


class Game(ABC, Generic[TState]):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def fullname(self) -> str:
        ...

    @property
    @abstractmethod
    def shape(self) -> tuple[int, int]:
        ...

    @abstractmethod
    def initial_state(self, start: str | list[int] | None = None) -> TState:
        ...

    @abstractmethod
    def config(self) -> GameParameters:
        ...

    @abstractmethod
    def from_kaggle(self, obs: Observation, config: Configuration) -> TState:
        ...

    def create_agent(self, player: Solver) -> Callable[[Observation, Configuration], int]:
        def get_best_move(obs: Observation, config: Configuration) -> int:
            state = self.from_kaggle(obs, config)
            move = player.solve(state)
            return move

        return get_best_move

    @abstractmethod
    def symmetries(
        self, encoded_state: np.ndarray, policy: np.ndarray
    ) -> list[tuple[np.ndarray, np.ndarray]]:
        ...

    @abstractmethod
    def display(self, state: TState) -> None:
        ...

    @abstractmethod
    def display_outcome(self, state: TState) -> None:
        ...

    def generate_states(
        self, depth: int, min_num_states: int = 1, start: str | list[int] | None = None
    ) -> list[TState]:
        state = self.initial_state(start)
        states = Game._get_all_states(state, depth)
        for _ in range(len(states), min_num_states):
            states.append(self.initial_state(start))

        return states

    @staticmethod
    def _get_all_states(state: TState, depth: int) -> list[TState]:
        all_states = [state]
        if depth == 0:
            return all_states

        legal_moves = state.status().legal_moves
        for move in legal_moves:
            new_state = state.play_move(move)
            states_at_next_depth = Game._get_all_states(new_state, depth - 1)
            all_states += states_at_next_depth

        return all_states


class ConnectX(Game[ConnectXState]):
    def __init__(self, rows: int = 6, cols: int = 7, view: View[ConnectXState] | None = None) -> None:
        self.rows = rows
        self.cols = cols
        self.view = view or ConsoleConnectXView()

    @property
    def name(self) -> str:
        return "connectx"

    @property
    def fullname(self) -> str:
        return f"ConnectX_{self.rows}x{self.cols}"

    @property
    def shape(self) -> tuple[int, int]:
        return self.rows, self.cols

    def initial_state(self, start: str | list[int] | None = None) -> ConnectXState:
        return ConnectXState.create(self.rows, self.cols, start)

    def config(self) -> GameParameters:
        NUM_PLANES = 3

        shape = self.shape
        observation_shape = NUM_PLANES, self.rows, self.cols
        action_size = self.cols
        return GameParameters(shape, observation_shape, action_size)

    def from_kaggle(self, obs: Observation, config: Configuration) -> ConnectXState:
        grid = np.asarray(obs.board).reshape(config.rows, config.columns)
        state = ConnectXState.from_grid(grid)
        return state

    def symmetries(
        self, encoded_state: np.ndarray, policy: np.ndarray
    ) -> list[tuple[np.ndarray, np.ndarray]]:
        syms: list[tuple[np.ndarray, np.ndarray]] = []

        flipped_state = encoded_state[:, :, ::-1]
        flipped_policy = policy[::-1]

        syms.append((encoded_state, policy))
        syms.append((flipped_state, flipped_policy))

        return syms

    def display(self, state: ConnectXState) -> None:
        self.view.display(state)

    def display_outcome(self, state: ConnectXState) -> None:
        self.view.display_outcome(state)


class TicTacToe(Game[TicTacToeState]):
    ROWS, COLS = 3, 3

    def __init__(self, view: View[TicTacToeState] | None = None) -> None:
        self.view = view or ConsoleTicTacToeView()

    @property
    def name(self) -> str:
        return "tictactoe"

    @property
    def fullname(self) -> str:
        return f"TicTacToe"

    @property
    def shape(self) -> tuple[int, int]:
        return self.ROWS, self.COLS

    def initial_state(self, start: str | list[int] | None = None) -> TicTacToeState:
        return TicTacToeState.create(start)

    def config(self) -> GameParameters:
        NUM_PLANES = 3

        shape = self.shape
        observation_shape = NUM_PLANES, self.ROWS, self.COLS
        action_size = self.ROWS * self.COLS
        return GameParameters(shape, observation_shape, action_size)

    def from_kaggle(self, obs: Observation, config: Configuration) -> TicTacToeState:
        grid = np.asarray(obs.board).reshape(self.ROWS, self.COLS)
        state = TicTacToeState.from_grid(grid)
        return state

    def symmetries(
        self, encoded_state: np.ndarray, policy: np.ndarray
    ) -> list[tuple[np.ndarray, np.ndarray]]:
        syms: list[tuple[np.ndarray, np.ndarray]] = []
        for rot in range(4):
            state = np.rot90(encoded_state, axes=(1, 2), k=rot)
            pol = np.rot90(policy.reshape(self.ROWS, self.COLS), k=rot).flatten()
            syms.append((state, pol))

        flipped_state = encoded_state[:, :, ::-1]
        flipped_policy = np.fliplr(policy.reshape(3, 3)).flatten()

        for rot in range(4):
            state = np.rot90(flipped_state, axes=(1, 2), k=rot)
            pol = np.rot90(flipped_policy.reshape(self.ROWS, self.COLS), k=rot).flatten()
            syms.append((state, pol))

        return syms

    def display(self, state: TicTacToeState) -> None:
        self.view.display(state)

    def display_outcome(self, state: TicTacToeState) -> None:
        self.view.display_outcome(state)


class Chess(Game[ChessState]):
    ROWS, COLS = 8, 8

    def __init__(self, view: View[ChessState] | None = None) -> None:
        self.view = view or ChessTicTacToeView()

    @property
    def name(self) -> str:
        return "chess"

    @property
    def fullname(self) -> str:
        return "Chess"

    @property
    def shape(self) -> tuple[int, int]:
        return self.ROWS, self.COLS

    def initial_state(self, start: str | list[int] | None = None) -> ChessState:
        return ChessState.create(start)

    def config(self) -> GameParameters:
        shape = self.shape
        return GameParameters(shape, ObsPlanes.shape(), ActionPlanes.size())

    def from_kaggle(self, obs: Observation, config: Configuration) -> ChessState:
        pass

    def symmetries(
        self, encoded_state: np.ndarray, policy: np.ndarray
    ) -> list[tuple[np.ndarray, np.ndarray]]:
        return [(encoded_state, policy)]

    def display(self, state: ChessState) -> None:
        self.view.display(state)

    def display_outcome(self, state: ChessState) -> None:
        self.view.display_outcome(state)


def get_game(game_type: GameType) -> Chess | ConnectX | TicTacToe:
    match game_type:
        case GameType.TICTACTOE:
            return TicTacToe()
        case GameType.CONNECTX:
            return ConnectX()
        case GameType.CHESS:
            return Chess()
        case _:
            raise ValueError(f"Invalid game type: {game_type}")



class BitboardUtil:
    __slots__ = ["rows", "cols", "num_slots", "BOTTOM_ROW", "BOARD_MASK"]

    def __init__(self, rows: int, cols: int):
        self.rows: int = rows
        self.cols: int = cols
        self.num_slots: int = rows * cols
        self.BOTTOM_ROW: int = self.get_bottom_row()
        self.BOARD_MASK: int = self.get_board_mask()

    def get_bottom_row(self) -> int:
        x = 1
        for _ in range(self.cols - 1):
            x |= x << self.rows
        return x

    def get_col_mask(self, col: int) -> int:
        first_col = (1 << (self.rows - 1)) - 1
        return first_col << (self.rows * col)

    def move_to_col(self, move: int) -> int:
        for i in range(self.cols):
            col_mask = self.get_col_mask(i)
            if col_mask & move:
                return i

        raise ValueError(f"Move {move} not associated with any column.")

    def get_board_mask(self) -> int:
        x = self.BOTTOM_ROW << (self.rows - 1)
        return x - self.BOTTOM_ROW

    def move_order(self) -> list[int]:
        order: list[int] = [0] * self.cols

        for i in range(self.cols):
            if i % 2 == 0:
                order[i] = (self.cols - i - 1) // 2
            else:
                order[i] = (self.cols + i) // 2

        return order

    def __repr__(self) -> str:
        return f"BitboardUtil(rows={self.rows}, cols={self.cols})"



import asyncio
import logging
from datetime import datetime

from caz.views.chess_main import Application


def launch() -> None:
    # Generate a timestamp string
    timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    logging.basicConfig(
        filename=f"app_{timestamp}.log",
        level=logging.INFO,
        format="%(asctime)s - %(levelname)s - %(message)s",
    )

    loop = asyncio.get_event_loop()
    app = Application(loop)
    asyncio.run(app.show_async())


if __name__ == "__main__":
    launch()



from .game import Chess as Chess
from .game import ConnectX as ConnectX
from .game import Game as Game
from .game import TicTacToe as TicTacToe
from .game import get_game as get_game
from .game_type import GameType as GameType



from __future__ import annotations

import asyncio
import logging
import tkinter as tk
from importlib import resources
from tkinter import filedialog, ttk

import chess
from PIL import Image, ImageTk

from ..alpha_zero import AlphaZero, PytorchNeuralNetwork
from ..games import Chess
from .chess_detail import DARK_COLOR, REFRESH_RATE, ChessScreen, ChessScreenController, ChessScreenModel

GREY_COLOR = "#BEC7C6"


class Application(tk.Tk):
    def __init__(self, event_loop):
        super().__init__()

        logging.info("CatchemAlphaZero app started.")

        game = Chess()
        self.loop = event_loop
        self.is_open = True
        self.title("CatchemAlphaZero")
        self.resizable(False, False)
        self.current_screen = None
        self.protocol("WM_DELETE_WINDOW", self.raise_exit_flag)

        logging.info("Loading Torch weights...")
        network = PytorchNeuralNetwork.create(game, ".")
        alpha_zero = AlphaZero(network)
        model = ChessScreenModel(alpha_zero, chess.BLACK)

        logging.info("Loading screens...")
        self.title_screen = TitleScreen(self, self.switch_to_game_screen)
        self.game_screen = ChessScreen(self, self.switch_to_title_screen)
        self.game_controller = ChessScreenController(model, self.game_screen)

        self.switch_to_title_screen()
        self.create_menu()

        logging.info("Loading icon...")
        with resources.path("caz.images", "tiny_chess.ico") as icon_path:
            self.iconbitmap(icon_path)

        logging.info("Initialization complete")

    async def show_async(self):
        while self.is_open:
            self.update()
            await asyncio.sleep(REFRESH_RATE)

    def raise_exit_flag(self):
        self.is_open = False

    def switch_to_title_screen(self):
        self.game_controller.model.stop_pondering()
        if self.current_screen:
            self.current_screen.pack_forget()
        self.title_screen.pack(fill=tk.BOTH, expand=True)
        self.current_screen = self.title_screen

    def switch_to_game_screen(self, human_colors: list[chess.Color]) -> None:
        if self.current_screen:
            self.current_screen.pack_forget()
        asyncio.create_task(self.game_controller.reset(human_colors))
        self.game_screen.pack(fill=tk.BOTH, expand=True)
        self.current_screen = self.game_screen

    def create_menu(self):
        menu_bar = tk.Menu(self)
        self.config(menu=menu_bar)

        file_menu = tk.Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label="File", menu=file_menu)

        file_menu.add_command(label="Open", command=self.open_file)
        file_menu.add_command(label="Exit", command=self.quit)

    def open_file(self):
        file_path = filedialog.askopenfilename()
        print("Selected file:", file_path)


class TitleScreen(tk.Frame):
    def __init__(self, master, switch_screen_callback):
        super().__init__(master)

        # Load and display the background image
        with resources.path("caz.images", "caz_splashscreen.jpg") as splash_path:
            logo = Image.open(splash_path).resize((614, 608))

        image = ImageTk.PhotoImage(logo)
        label = tk.Label(self, image=image)
        label.image = image  # avoid garbage collection
        label.pack(fill=tk.BOTH, expand=True)

        # Create a custom style for the buttons
        style = ttk.Style()
        style.configure(
            "TButton",
            foreground="#545454",
            background="white",
            font=("Helvetica", 14),
            padding=8,
            relief=tk.FLAT,
        )

        # Create a custom style for the buttons
        style = ttk.Style()
        style.theme_use("clam")  # put the theme name here, that you want to use
        style.configure(
            "TButton",
            foreground=DARK_COLOR,
            background=GREY_COLOR,
            font=("Cascadia Mono", 12),
            relief=tk.FLAT,
            anchor=tk.CENTER,
        )

        # Create the "Go to Title Screen" button
        self.white_button = ttk.Button(
            self,
            text="Play as White",
            command=lambda: switch_screen_callback([chess.WHITE]),
            style="TButton",
            compound=tk.CENTER,
        )

        self.black_button = ttk.Button(
            self,
            text="Play as Black",
            command=lambda: switch_screen_callback([chess.BLACK]),
            style="TButton",
            compound=tk.CENTER,
        )

        self.both_button = ttk.Button(
            self,
            text="Play as Both ",
            command=lambda: switch_screen_callback([chess.WHITE, chess.BLACK]),
            style="TButton",
            compound=tk.CENTER,
        )

        self.spectate_button = ttk.Button(
            self,
            text="Spectate Mode",
            command=lambda: switch_screen_callback([]),
            style="TButton",
            compound=tk.CENTER,
        )

        self.white_button.place(relx=0.38, rely=0.80, anchor=tk.CENTER)
        self.black_button.place(relx=0.38, rely=0.88, anchor=tk.CENTER)
        self.both_button.place(relx=0.62, rely=0.80, anchor=tk.CENTER)
        self.spectate_button.place(relx=0.62, rely=0.88, anchor=tk.CENTER)


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    app = Application(loop)
    asyncio.run(app.show_async())



from enum import Enum

from .solver import Solver


class AgentType(Enum):
    """Enum representing the types of agents supported."""

    HUMAN = "HUMAN"
    MINIMAX = "MINIMAX"
    MCTS = "MCTS"
    AZ = "AZ"

    @staticmethod
    def from_str(value: str) -> "AgentType":
        """Converts a string to its enum representation.

        Args:
            value (str): The string representation of the agent type.

        Raises:
            ValueError: If the string is not recognised.

        Returns:
            AgentType: The enum
        """
        if value.upper() == "HUMAN":
            return AgentType.HUMAN
        if value.upper() == "MINIMAX":
            return AgentType.MINIMAX
        if value.upper() == "MCTS":
            return AgentType.MCTS
        if value.upper() == "AZ":
            return AgentType.AZ
        supported_types = ", ".join([e.name for e in AgentType])
        message = f"{value} not a supported solver type. Supported types are {supported_types}."
        raise ValueError(message)

    def create_player(self, game) -> Solver:
        match self:
            case AgentType.HUMAN:
                raise ValueError("todo")
            case AgentType.MINIMAX:
                raise ValueError("todo")
            case AgentType.MCTS:
                from .mcts import MctsSolver

                return MctsSolver()
            case AgentType.AZ:
                from ..alpha_zero import AlphaZero, PytorchNeuralNetwork

                neural_net = PytorchNeuralNetwork.create(game, ".")
                alpha_zero = AlphaZero(neural_net)
                return alpha_zero.as_solver(num_mcts_sims=100)
            case _:
                raise ValueError(f"Unsupported agent type: {self}")



from dataclasses import dataclass, field
from typing import Self

from ..states.state import TemperatureSchedule


@dataclass
class AZMctsParameters:
    num_mcts_sims: int
    dirichlet_epsilon: float
    dirichlet_alpha: float
    discount_factor: float

    @classmethod
    def defaults(cls, fullname: str) -> Self:
        match fullname:
            case "TicTacToe":
                params = {
                    "num_mcts_sims": 500,
                    "dirichlet_epsilon": 0.25,
                    "dirichlet_alpha": 0.9,
                    "discount_factor": 0.99,
                }

            case "ConnectX_6x7":
                params = {
                    "num_mcts_sims": 600,
                    "dirichlet_epsilon": 0.25,
                    "dirichlet_alpha": 0.6,
                    "discount_factor": 0.98,
                }

            case "Chess":
                params = {
                    "num_mcts_sims": 800,
                    "dirichlet_epsilon": 0.25,
                    "dirichlet_alpha": 0.3,
                    "discount_factor": 0.99,
                }

            case _:
                raise NotImplementedError(f"MCTS defaults for {fullname} not supported.")

        return cls(**params)


@dataclass
class AZArenaParameters:
    num_games: int = 400
    num_mcts_sims: int = 60
    required_win_ratio: float = 0.55
    temperature_schedule: TemperatureSchedule = field(
        default_factory=lambda: TemperatureSchedule(4, 1.25)
    )

    @classmethod
    def defaults(cls, fullname: str) -> Self:
        match fullname:
            case "TicTacToe":
                params = {
                    "num_mcts_sims": 500,
                    "dirichlet_epsilon": 0.25,
                    "dirichlet_alpha": 0.9,
                    "discount_factor": 0.99,
                }

            case "ConnectX_6x7":
                params = {
                    "num_mcts_sims": 1600,
                    "dirichlet_epsilon": 0.25,
                    "dirichlet_alpha": 0.6,
                    "discount_factor": 0.98,
                }

            case _:
                raise NotImplementedError(f"Arena defaults for {fullname} not supported.")

        return cls(**params)


@dataclass
class AZTrainingParameters:
    num_generations: int
    num_epochs: int
    games_per_generation: int
    num_parallel: int
    minibatch_size: int
    temperature: TemperatureSchedule
    mcts_parameters: AZMctsParameters
    arena_parameters: AZArenaParameters

    @classmethod
    def defaults(cls, fullname: str) -> Self:
        mcts_parameters = AZMctsParameters.defaults(fullname)
        arena_parameters = AZArenaParameters()

        match fullname:
            case "TicTacToe":
                params = {
                    "num_generations": 20,
                    "num_epochs": 4,
                    "games_per_generation": 100,
                    "num_parallel": 50,
                    "minibatch_size": 128,
                    "temperature": TemperatureSchedule(2, 1.25),
                    "mcts_parameters": mcts_parameters,
                    "arena_parameters": arena_parameters,
                }

            case "ConnectX_6x7":
                params = {
                    "num_generations": 50,
                    "num_epochs": 4,
                    "games_per_generation": 400,
                    "num_parallel": 100,
                    "minibatch_size": 512,
                    "temperature": TemperatureSchedule(10, 1.2),
                    "mcts_parameters": mcts_parameters,
                    "arena_parameters": arena_parameters,
                }

            case "Chess":
                params = {
                    "num_generations": 500,
                    "num_epochs": 3,
                    "games_per_generation": 400,
                    "num_parallel": 100,
                    "minibatch_size": 1024,
                    "temperature": TemperatureSchedule(20, 1.2),
                    "mcts_parameters": mcts_parameters,
                    "arena_parameters": arena_parameters,
                }

            case _:
                raise NotImplementedError(f"Training defaults for {fullname} not supported.")

        return cls(**params)


@dataclass
class AZNetworkParameters:
    num_resnet_blocks: int
    num_features: int
    optimizer_learn_rate: float
    optimizer_weight_decay: float

    @classmethod
    def defaults(cls, fullname: str) -> Self:
        match fullname:
            case "TicTacToe":
                params = {
                    "num_resnet_blocks": 4,
                    "num_features": 64,
                    "optimizer_learn_rate": 0.001,
                    "optimizer_weight_decay": 0.0001,
                }

            case "ConnectX_6x7":
                params = {
                    "num_resnet_blocks": 9,
                    "num_features": 128,
                    "optimizer_learn_rate": 0.001,
                    "optimizer_weight_decay": 0.0001,
                }

            case "Chess":
                params = {
                    "num_resnet_blocks": 19,
                    "num_features": 256,
                    "optimizer_learn_rate": 0.0001,
                    "optimizer_weight_decay": 0.0001,
                }

            case _:
                raise NotImplementedError(f"Network defaults for {fullname} not supported.")

        return cls(**params)



import numpy as np
import pytest

from caz.games import Chess
from caz.states.chess import ObsPlanes
from caz.states.chess_enums import KeyGames


class TestChessFeatures:
    @pytest.mark.parametrize(
        "label,turn,can_player_kingside,can_player_queenside,can_opp_kingside,can_opp_queenside,half_move_clock,ep_square",
        [
            ("white_en_passant_sans", 1, 1, 1, 1, 1, 0, 1),
            ("black_en_passant_sans", 0, 1, 1, 1, 1, 0, 1),
            ("white_no_queenside_sans", 1, 1, 0, 1, 1, 0, 1),
            ("white_no_kingside_sans", 0, 1, 1, 0, 1, 1, 0),
            ("white_no_castling_sans", 1, 0, 0, 1, 1, 0, 1),
            ("black_no_queenside_sans", 0, 1, 0, 1, 1, 4, 0),
            ("black_no_kingside_sans", 0, 0, 1, 1, 1, 4, 0),
            ("black_no_castling_sans", 1, 1, 1, 0, 0, 4, 0),
            ("white_castled_black_not_yet_sans", 1, 0, 0, 1, 1, 0, 0),
        ],
    )
    def test_feature_creation(
        self,
        label: str,
        turn: float,
        can_player_kingside: float,
        can_player_queenside: float,
        can_opp_kingside: float,
        can_opp_queenside: float,
        half_move_clock: float,
        ep_square: float | None,
    ) -> None:
        # Arrange
        game = Chess()
        sans = KeyGames.get(label)
        sut = game.initial_state(sans)

        # Act
        actual = sut.to_feature()

        # Assert
        assert np.all(actual[ObsPlanes.TURN, :, :] == turn)
        assert np.all(actual[ObsPlanes.CAN_PLAYER_KINGSIDE, :, :] == can_player_kingside)
        assert np.all(actual[ObsPlanes.CAN_PLAYER_QUEENSIDE, :, :] == can_player_queenside)
        assert np.all(actual[ObsPlanes.CAN_OPP_KINGSIDE, :, :] == can_opp_kingside)
        assert np.all(actual[ObsPlanes.CAN_OPP_QUEENSIDE, :, :] == can_opp_queenside)
        assert np.all(actual[ObsPlanes.HALFMOVE_CLOCK, :, :] == half_move_clock)
        assert np.sum(actual[ObsPlanes.EN_PASSANT_SQ, :, :]) == ep_square



from __future__ import annotations

import random
from copy import copy
from dataclasses import dataclass
from math import log, sqrt
from typing import Generic

from ..states import State, Status, TMove, TState
from . import Solver


@dataclass
class MctsSolver(Solver):
    num_mcts_sims: int = 1_000

    def solve(self, state: State[TMove]) -> TMove:
        root = self.search(state)
        max_child = max(root.children, key=lambda c: c.visit_count)
        return max_child.move

    def search(self, root_state: State[TMove]) -> Node:
        root: Node[TMove] = Node(root_state.status(), None, None)

        for _ in range(self.num_mcts_sims):
            node = root
            state = copy(root_state)

            # Selection
            while not node.is_leaf_node and node.has_legal_moves:
                node = node.select_child()
                state.set_move(node.move)

            # Expansion
            if node.unexplored_moves:
                move = random.choice(node.unexplored_moves)
                state.set_move(move)
                child_node = Node(state.status(), parent=node, move=move)
                node.unexplored_moves.remove(move)
                node.children.append(child_node)
                node = child_node

            # Simulation (aka rollout)
            while (status := state.status()).is_in_progress:
                move = random.choice(status.legal_moves)
                state.set_move(move)

            # Backpropagate
            while node:
                outcome = status.outcome(node.played_by)
                node.update(outcome)
                node = node.parent

        return root


class Node(Generic[TMove]):
    __slots__ = ["move", "parent", "played_by", "wins", "visit_count", "children", "unexplored_moves"]

    def __init__(
        self, status: Status[TMove], parent: Node[TMove] | None = None, move: TMove | None = None
    ):
        self.move = move
        self.parent = parent
        self.played_by = status.played_by

        self.wins: int = 0
        self.visit_count: int = 0

        self.children: list[Node[TMove]] = []
        self.unexplored_moves: list[TMove] = status.legal_moves

    @property
    def has_legal_moves(self) -> bool:
        return bool(self.children or self.unexplored_moves)

    @property
    def is_leaf_node(self) -> bool:
        return bool(self.unexplored_moves)

    def update(self, outcome: int) -> None:
        self.visit_count += 1
        self.wins += (1.0 + outcome) / 2.0

    def select_child(self) -> "Node[TMove]":
        return max(self.children, key=lambda c: c.ucb())

    def ucb(self) -> float:
        c = sqrt(2)
        exploitation_param = self.wins / self.visit_count
        exploration_param = sqrt(log(self.parent.visit_count) / self.visit_count)
        return exploitation_param + c * exploration_param

    def html(self, state: TState) -> str:
        from ..views.html import MctsNodeHtmlBuilder

        html_printer = MctsNodeHtmlBuilder()
        return html_printer.build_tiny_html(self, state)

    def render(self, state: State[TMove]):
        from ..graph import visualize_tree

        return visualize_tree(state, self)

    def __repr__(self):
        return f"Node(move={self.move}, W/V={self.wins}/{self.visit_count}, ({len(self.unexplored_moves)} unexplored moves))"



import pickle
from pathlib import Path

from ..games import ConnectX, Game
from .alpha_zero_parameters import AZTrainingParameters
from .network import TrainingData
from .network_torch import PytorchNeuralNetwork


def learn_from_data(
    generation: int, training_data: list[TrainingData], training_params: AZTrainingParameters
):
    neural_net.train(training_data, training_params.num_epochs, training_params.minibatch_size)
    neural_net.generation = generation + 1
    neural_net.save()


def get_training_files(game: Game, directory: Path | str) -> None:
    directory = Path(directory)

    training_regex = f"{game.fullname}_gen*.pkl"
    training_files = [file_path for file_path in directory.glob(training_regex)]

    for training_file in training_files:
        generation = int(training_file.stem.split("gen")[1])
        with open(training_file, "rb") as f:
            data = pickle.load(f)

        yield (generation, data)


if __name__ == "__main__":
    game = ConnectX()
    training_params = AZTrainingParameters.defaults(game.fullname)
    neural_net = PytorchNeuralNetwork.create(game, "..")

    for gen, data in get_training_files(game, r"../training_data"):
        print(gen)
        learn_from_data(gen, data, training_params)



from colorama import Fore, Style

from ..states import ChessState, ConnectXState, TicTacToeState
from . import View


class ConsoleConnectXView(View[ConnectXState]):
    def display(self, state: ConnectXState) -> None:
        grid = state.to_grid()
        num_rows, num_cols = grid.shape

        color_map = {0: Fore.WHITE, 1: Fore.RED, 2: Fore.YELLOW}

        # Build the string representation
        output = ""
        for row in range(num_rows):
            for col in range(num_cols):
                cell = grid[row, col]
                color = color_map.get(cell, Fore.RESET)
                output += f"{color}{cell} "
            output += "\n"

        output += Style.RESET_ALL

        print(output)

    def display_outcome(self, state: ConnectXState) -> None:
        if state.is_won():
            print(f"Player {state.played_by}  wins!")
        else:
            print("It's a draw!")


class ConsoleTicTacToeView(View[TicTacToeState]):
    def display(self, state: TicTacToeState) -> None:
        grid = state.to_grid()
        print(grid)

    def display_outcome(self, state: TicTacToeState) -> None:
        if state.is_won():
            print(f"Player {state.played_by}  wins!")
        else:
            print("It's a draw!")


class ChessTicTacToeView(View[ChessState]):
    def display(self, state: ChessState) -> None:
        print(state.board)
        print()

    def display_outcome(self, state: ChessState) -> None:
        if state.status().value:
            print(f"Player {state.played_by}  wins!")
        else:
            print("It's a draw!")



from dataclasses import dataclass, field

import numpy as np

from ..states import State


@dataclass(slots=True)
class RecordedAction:
    state_before: State
    policy: np.ndarray
    move: int
    state_after: State


@dataclass(slots=True)
class ParallelGame:
    initial_state: State
    recorded_actions: list[RecordedAction] = field(default_factory=list)

    @property
    def latest_state(self) -> State:
        if self.recorded_actions:
            return self.recorded_actions[-1].state_after
        return self.initial_state



import sys
from argparse import ArgumentParser, Namespace
from typing import Sequence

import numpy as np

from .games import GameType, get_game
from .solvers import AgentType, Solver


def run(args: Namespace) -> None:
    game_type: GameType = args.game
    init: str = args.init
    player1_type: AgentType = args.player1
    player2_type: AgentType = args.player2

    game = get_game(game_type)
    state = game.initial_state(init)

    player1 = player1_type.create_player(game)
    player2 = player2_type.create_player(game)

    player, opponent = player1, player2
    while state.status().is_in_progress:
        move = player.solve(state)
        state.set_move(move)
        game.display(state)
        player, opponent = opponent, player

    game.display_outcome(state)


def kaggle(args: Namespace) -> None:
    from kaggle_environments import make

    game_type: GameType = args.game
    player1_type: AgentType = args.player1
    player2_type: AgentType = args.player2

    game = get_game(game_type)

    player1 = player1_type.create_player(game)
    player2 = player2_type.create_player(game)
    agent1 = game.create_agent(player1)
    agent2 = game.create_agent(player2)

    # Setup a ConnectX environment.
    env = make(game.name, debug=True)
    env.run([agent1, agent2])
    env.render(mode="ipython")


def learn(args: Namespace) -> None:
    from .alpha_zero import AlphaZero, AZTrainingParameters, PytorchNeuralNetwork

    game_type: GameType = args.game
    init: str = args.init

    game = get_game(game_type)

    # Build alpha zero with latest weights
    neural_net = PytorchNeuralNetwork.create(game, ".")
    alpha_zero = AlphaZero(neural_net)

    training_params = AZTrainingParameters.defaults(game.fullname)
    alpha_zero.self_learn(training_params, init)

    print("done!")


def supervised_learning(args: Namespace) -> None:
    import chess

    from .alpha_zero import AZTrainingParameters, PytorchNeuralNetwork, SupervisedTrainer

    init: str = args.init

    game = get_game(GameType.CHESS)
    training_params = AZTrainingParameters.defaults(game.fullname)

    # Build alpha zero with latest weights
    neural_net = PytorchNeuralNetwork.create(game, ".")

    path = r".\stockfish\stockfish-windows-2022-x86-64-avx2.exe"
    with chess.engine.SimpleEngine.popen_uci(path) as stockfish:
        trainer = SupervisedTrainer(neural_net, stockfish)
        trainer.train(training_params)

    print("done!")


def hyper(args: Namespace) -> None:
    from .alpha_zero import AlphaZero, AZArenaParameters, PytorchNeuralNetwork

    generation: int = args.gen
    game_type: GameType = args.game
    game = get_game(game_type)

    # Build alpha zero with latest weights
    neural_net = PytorchNeuralNetwork.create(game, ".", generation)
    alpha_zero = AlphaZero(neural_net)
    arena_params = AZArenaParameters()

    alpha_zero.explore_hyperparameters(generation, arena_params)

    print("done!")


def main() -> None:
    parse_args(sys.argv[1:])


def parse_args(args: Sequence[str]) -> None:
    parser = ArgumentParser()
    subparsers = parser.add_subparsers()

    # Runs an adversarial game
    run_parser = subparsers.add_parser("run")
    run_parser.add_argument("--game", type=GameType.from_str, default=GameType.TICTACTOE)
    run_parser.add_argument("--player1", type=AgentType.from_str, default=AgentType.HUMAN)
    run_parser.add_argument("--player2", type=AgentType.from_str, default=AgentType.HUMAN)
    run_parser.add_argument("--init", type=str, default="")
    run_parser.set_defaults(func=run)

    # Runs an adversarial game on the kaggle platform
    run_parser = subparsers.add_parser("kaggle")
    run_parser.add_argument("--game", type=GameType.from_str, default=GameType.TICTACTOE)
    run_parser.add_argument("--player1", type=AgentType.from_str, default=AgentType.HUMAN)
    run_parser.add_argument("--player2", type=AgentType.from_str, default=AgentType.HUMAN)
    run_parser.set_defaults(func=kaggle)

    # Trains a deep neural net via reinforcement learning
    run_parser = subparsers.add_parser("learn")
    run_parser.add_argument("--game", type=GameType.from_str, default=GameType.TICTACTOE)
    run_parser.add_argument("--init", type=str, default=None)
    run_parser.set_defaults(func=learn)

    # Trains a deep neural net via supervised learning
    run_parser = subparsers.add_parser("super")
    run_parser.add_argument("--init", type=str, default=None)
    run_parser.set_defaults(func=supervised_learning)

    # Explores learning rate as a function of training hyperparameters
    run_parser = subparsers.add_parser("hyper")
    run_parser.add_argument("--game", type=GameType.from_str, default=GameType.TICTACTOE)
    run_parser.add_argument("--gen", type=int, default=0)
    run_parser.set_defaults(func=hyper)

    namespace = parser.parse_args(args)
    namespace.func(namespace)



from .chess import ChessState as ChessState
from .connectx import ConnectXState as ConnectXState
from .state import State as State
from .state import Status as Status
from .state import TMove as TMove
from .state import TState as TState
from .tictactoe import TicTacToeState as TicTacToeState



from copy import copy

from ..states import ConnectXState


class Solver:
    def minimax(self, state: ConnectXState, alpha: int, beta: int) -> int:
        if state.is_full():
            return 0

        win_mask = state.win_mask()
        possible_moves = state._possible_bitmoves_mask()
        if win_mask & possible_moves:
            return (state.num_slots - state.num_moves + 1) // 2

        max_possible_score = (state.num_slots - state.num_moves - 1) // 2
        if max_possible_score <= alpha:
            return max_possible_score

        alpha = -100_000_000
        beta = min(beta, max_possible_score)

        for move in state.possible_bitmoves():
            b = copy(state)
            b.set_bitmove(move)
            score = -self.minimax(b, -beta, -alpha)
            alpha = max(alpha, score)
            if score >= beta:
                return alpha

        return alpha



from dataclasses import dataclass
from typing import Protocol

import numpy as np
from numpy.typing import NDArray

from ..games import Game
from ..states import State, TMove


@dataclass
class TrainingData:
    encoded_state: np.ndarray
    policy: np.ndarray
    outcome: float


class NeuralNetwork(Protocol):
    @property
    def game(self) -> Game:
        ...

    @property
    def action_size(self) -> int:
        ...

    @property
    def generation(self) -> int:
        ...

    def predict(self, state: State[TMove]) -> tuple[NDArray[np.float32], float]:
        ...

    def predict_parallel(self, state: list[State]) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
        ...

    def set_to_eval(self) -> None:
        ...

    def train(
        self,
        training_set: list[TrainingData],
        num_epochs: int,
        minibatch_size: int,
        log_progress: bool = True,
    ) -> None:
        ...

    def save_training_data(self, training_set: list[TrainingData]) -> None:
        ...

    def load_training_data(self, int: generation) -> list[TrainingData]:
        ...

    def save(self) -> None:
        ...


class DummyNeuralNetwork:
    def predict(self, state: State[TMove]) -> tuple[NDArray[np.float32], float]:
        policy: NDArray[np.float32] = np.ones(state.action_size, dtype=np.float32)
        return policy / policy.sum(), 0.0

    def set_to_eval(self) -> None:
        pass

    def train(self, training_set: list[TrainingData]) -> None:
        ...

    def save(self, generation: int) -> None:
        ...



from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray

from ..states import ConnectXState, State, TicTacToeState
from ..states.state import Policy


@dataclass
class PoliciedState:
    state: State
    policy_grid: NDArray

    def _repr_html_(self) -> str:
        return self.state.html(self.policy_grid, False)


def show_tictactoe_policy(
    self: TicTacToeState, policy: Policy, include_chart: bool = True
) -> PoliciedState:
    encoded_policy = policy.encoded_policy
    q_value = policy.value

    if include_chart:
        _, ax = plt.subplots(figsize=(5, 2))
        ax.spines["top"].set_edgecolor("white")
        ax.spines["right"].set_edgecolor("white")
        ax.spines["bottom"].set_edgecolor("black")
        ax.spines["left"].set_edgecolor("white")
        ax.set_yticks([])
        ax.set_xlabel(f"v ∈ (-1, +1) = {q_value:.2}")

        plt.bar(["a3", "b3", "c3", "a2", "b2", "c2", "a1", "b1", "c1"], encoded_policy, axes=ax)

    policy_grid = encoded_policy.reshape((self.rows, self.cols))
    return PoliciedState(self, policy_grid)


def show_connectx_policy(
    self: ConnectXState, policy: Policy, include_chart: bool = True
) -> PoliciedState:
    encoded_policy = policy.encoded_policy
    q_value = policy.value

    if include_chart:
        _, ax = plt.subplots(figsize=(4.34, 2))
        plt.bar(range(1, self.cols + 1), encoded_policy, axes=ax)

        ax.spines["top"].set_edgecolor("white")
        ax.spines["right"].set_edgecolor("white")
        ax.spines["bottom"].set_edgecolor("black")
        ax.spines["left"].set_edgecolor("white")
        ax.set_yticks([])
        ax.set_xticks([])
        ax.set_xlabel(f"v ∈ (-1, +1) = {q_value:.2}")

    policy_grid = np.tile(encoded_policy, (self.rows, 1))
    return PoliciedState(self, policy_grid)


ConnectXState.show_policy = show_connectx_policy
TicTacToeState.show_policy = show_tictactoe_policy



from dataclasses import dataclass
from typing import Self

import numpy as np

from ..bitboard_utils import BitboardUtil
from .state import State, Status, TemperatureSchedule

# 3  7 11
# 2  6 10
# 1  5  9
# 0  4  8

MOVES = [1 << 2, 1 << 6, 1 << 10, 1 << 1, 1 << 5, 1 << 9, 1 << 0, 1 << 4, 1 << 8]
POWERS = np.array(MOVES).reshape(3, 3)


@dataclass(slots=True)
class TicTacToeState(State[int]):
    bitboard_util: BitboardUtil
    mask: int
    position: int
    num_moves: int

    @property
    def rows(self) -> int:
        return 3

    @property
    def cols(self) -> int:
        return 3

    @property
    def shape(self) -> tuple[int, int]:
        return self.rows, self.cols

    @property
    def num_slots(self) -> int:
        return self.rows * self.cols

    @property
    def is_full(self) -> bool:
        return self.num_moves == self.num_slots

    @property
    def played_by(self) -> int:
        is_odd_num_moves = self.num_moves & 1
        return 2 - is_odd_num_moves

    @property
    def move_count(self) -> int:
        return self.num_moves

    def status(self) -> Status[int]:
        is_won = self.is_won()
        is_ended = is_won or self.is_full

        is_in_progress = not is_ended
        value = 1.0 if is_won else 0.0
        legal_moves = [] if is_ended else self._possible_moves_unchecked()
        return Status(is_in_progress, self.played_by, value, legal_moves)

    def play_move(self, move: int) -> Self:
        state = self.__copy__()
        state.set_move(move)
        return state

    def set_move(self, move: int) -> None:
        bitmove = MOVES[move]
        self.position ^= self.mask
        self.mask |= bitmove
        self.num_moves += 1

    def policy_loc(self, move: int) -> int:
        return move

    def is_won(self) -> bool:
        rows = self.rows + 1
        directions = (1, rows - 1, rows, rows + 1)
        bitboard = self.position ^ self.mask
        for dir in directions:
            if bitboard & (bitboard >> dir) & (bitboard >> 2 * dir):
                return True

        return False

    def to_feature(self) -> np.ndarray:
        player_to_move = self.position
        opponent_of_player_to_move = self.position ^ self.mask
        r = np.sign(player_to_move & POWERS, dtype=np.float32)
        g = np.sign(opponent_of_player_to_move & POWERS, dtype=np.float32)
        b = 1 - r - g
        return np.stack((r, g, b))

    def to_grid(self) -> np.ndarray:
        posn = self.position ^ self.mask if self.num_moves & 1 else self.position
        player_1 = posn
        player_2 = posn ^ self.mask
        r = np.sign(player_1 & POWERS)
        g = np.sign(player_2 & POWERS)
        return np.asarray(r + 2 * g, dtype=np.int8)

    def get_input_move(self) -> int:
        while True:
            response = input("Your turn to move. Please enter an integer: ")
            try:
                move = int(response)
                if move in self.status().legal_moves:
                    return move
            except ValueError:
                pass
            print("Move was invalid. Please try again.")

    def html(self, policy: np.ndarray | None = None, is_tiny_repr: bool = False) -> str:
        from ..views.html import TicTacToeHtmlBuilder

        html_printer = TicTacToeHtmlBuilder()

        if is_tiny_repr:
            return html_printer.build_tiny_html(self)

        return html_printer.build_html(self, policy)

    def plot(self):
        from ..views.plot import plot_state

        plot_state(self.to_feature(), figsize=(3, 2), linewidth=1.5)

    def _possible_moves_unchecked(self) -> list[int]:
        possible_moves_mask = self._possible_bitmoves_mask()
        return [i for i, move in enumerate(MOVES) if possible_moves_mask & move]

    def _possible_bitmoves_mask(self) -> int:
        return self.mask ^ self.bitboard_util.BOARD_MASK

    def _repr_html_(self) -> str:
        return self.html()

    def __copy__(self) -> "TicTacToeState":
        return TicTacToeState(self.bitboard_util, self.mask, self.position, self.num_moves)

    @classmethod
    def from_grid(cls, grid: np.ndarray) -> "TicTacToeState":
        rows, cols = grid.shape
        padded_grid = np.vstack((np.zeros(cols), grid))

        indices = np.flipud(np.arange((rows + 1) * cols).reshape((cols, rows + 1)).transpose())
        binary_vals = 2 ** indices.astype(np.int64)

        mask = (padded_grid > 0).astype(np.int64)
        num_moves = np.sum(mask)
        mark = 1 + num_moves % 2

        posn = (padded_grid == mark).astype(np.int64)
        mask_utils = BitboardUtil(rows + 1, cols)
        board = cls(mask_utils, 0, 0, 0)
        mask_val = np.sum(mask * binary_vals)
        posn_val = np.sum(posn * binary_vals)
        board.mask = mask_val
        board.position = posn_val
        board.num_moves = int(np.sum(mask))
        return board

    @classmethod
    def create(cls, moves: str | list[int] | None = None) -> "TicTacToeState":
        mask = BitboardUtil(3 + 1, 3)
        state = cls(mask, 0, 0, 0)
        moves = moves or []

        if isinstance(moves, str):
            move_array = moves.replace(" ", "").split(",")
            moves = [int(move) for move in move_array]

        for move in moves:
            state.set_move(move)
        return state



import math
import random

import numpy as np
from chess import Move
from chess.engine import Limit, PovScore, SimpleEngine
from tqdm import trange

from ..states.chess import ChessState
from .alpha_zero_parameters import AZTrainingParameters
from .network import NeuralNetwork, TrainingData


class SupervisedTrainer:
    def __init__(self, neural_net: NeuralNetwork, engine: SimpleEngine) -> None:
        self.neural_net = neural_net
        self.engine = engine

    def train(self, training_params: AZTrainingParameters) -> None:
        discount_factor = training_params.mcts_parameters.discount_factor
        num_epochs, minibatch_size = training_params.num_epochs, training_params.minibatch_size

        for i in trange(training_params.num_generations):
            training_data = self.generate_training_data(10_000, discount_factor)

            extended_training_set = self._exploit_symmetries(training_data)

            # Train against the newly generated games
            self.neural_net.train(extended_training_set, num_epochs, minibatch_size)
            self.neural_net.generation += 1

            if i % 10 == 0:
                self.neural_net.save_training_data(extended_training_set)
                self.neural_net.save()

    def _exploit_symmetries(self, training_data) -> list[TrainingData]:
        # todo
        return training_data

    def generate_training_data(self, min_num_points: int, discount_factor: float):
        BLUNDER_PROBABILITY = 0.2

        training_set: list[TrainingData] = []
        while len(training_set) < min_num_points:
            state: ChessState = self.neural_net.game.initial_state()
            status = state.status()
            while status.is_in_progress:
                training_data, top_moves = self.to_data_point(state, discount_factor)
                training_set.append(training_data)

                if random.random() < BLUNDER_PROBABILITY:
                    legal_moves = status.legal_moves
                    idx = np.random.choice(len(legal_moves))
                    move = legal_moves[idx]
                    state.set_move(move)
                else:
                    idx = np.random.choice(len(top_moves))
                    top_move = top_moves[idx]
                    state.set_move(top_move)

                status = state.status()

        return training_set

    def build_policy(
        self, state: ChessState, top_moves: list[Move], top_scores: np.ndarray, noise: float
    ) -> np.ndarray:
        action_size = self.neural_net.game.config().action_size
        encoded_policy = np.zeros(action_size, dtype=np.float32)

        priors = SupervisedTrainer.policy_weights(top_scores)
        for move, prior in zip(top_moves, priors):
            loc = state.policy_loc(move)
            encoded_policy[loc] = prior

        for move in state.status().legal_moves:
            loc = state.policy_loc(move)
            encoded_policy[loc] += noise

        encoded_policy /= encoded_policy.sum()
        return encoded_policy

    def to_data_point(
        self, state: ChessState, discount_factor: float
    ) -> tuple[TrainingData, list[Move]]:
        encoded_state = state.to_feature()

        time_limit = Limit(time=0.01)
        infos = self.engine.analyse(state.board, time_limit, multipv=5)

        top_moves = [info["pv"][0] for info in infos]
        top_scores = np.array([SupervisedTrainer.score_to_centipawns(info["score"]) for info in infos])
        score = top_scores.max()
        outcome = SupervisedTrainer.centipawns_to_q_value(score)

        NOISE = 0.001
        policy = self.build_policy(state, top_moves, top_scores, NOISE)
        training_data = TrainingData(encoded_state, policy, outcome)
        return training_data, top_moves

    @staticmethod
    def score_to_centipawns(score: PovScore):
        if score.is_mate():
            full_moves = 2 * np.abs(score.relative.moves) - (1 if score.relative.moves > 0 else 0)
            value = 0.99**full_moves
            signed_value = np.sign(score.relative.moves) * value
            return SupervisedTrainer.q_value_to_centipawns(signed_value)
        return score.relative.score()

    @staticmethod
    def policy_weights(scores: np.ndarray) -> np.ndarray:
        drops = (scores.max() - scores) / 100.0
        policies = 0.15**drops
        return policies

    @staticmethod
    def centipawns_to_q_value(cp: int) -> float:
        return math.atan(cp / 250) / 1.5620688421

    @staticmethod
    def q_value_to_centipawns(q: float) -> int:
        cp = 250 * math.tan(1.5620688421 * q)
        return round(cp)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Self, TypeVar

import numpy as np

TMove = TypeVar("TMove")
TState = TypeVar("TState", bound="State[+TMove]")
TState_co = TypeVar("TState_co", bound="State[+TMove]", covariant=True)


class State(ABC, Generic[TMove]):
    @property
    @abstractmethod
    def played_by(self) -> int:
        ...

    @property
    @abstractmethod
    def move_count(self) -> int:
        ...

    @property
    @abstractmethod
    def shape(self) -> tuple[int, int]:
        ...

    @abstractmethod
    def status(self) -> Status[TMove]:
        ...

    @abstractmethod
    def play_move(self, move: TMove) -> Self:
        ...

    @abstractmethod
    def set_move(self, move: TMove) -> None:
        ...

    @abstractmethod
    def policy_loc(self, move: TMove) -> int:
        ...

    @abstractmethod
    def get_input_move(self) -> TMove:
        ...

    @abstractmethod
    def html(self, is_tiny_repr: bool = False) -> str:
        ...

    @abstractmethod
    def to_feature(self) -> np.ndarray:
        ...

    @abstractmethod
    def __copy__(self) -> Self:
        ...


@dataclass
class Policy(Generic[TMove]):
    moves: list[TMove] | None
    priors: np.ndarray[np.float32] | None
    encoded_policy: np.ndarray[np.float32]
    value: float
    node: "Node" | None = None

    def select_move(self, temperature: float) -> TMove:
        if temperature <= 0.001:
            idx = np.argmax(self.priors)
        else:
            temperature_policy = self.priors ** (1 / temperature)
            temperature_policy /= temperature_policy.sum()
            idx = np.random.choice(len(self.priors), p=temperature_policy)

        return self.moves[idx]


@dataclass
class TemperatureSchedule:
    cutoff: int
    temperature: float

    def get_temperature(self, move_count: int) -> float:
        return self.temperature if move_count < self.cutoff else 0

    @classmethod
    def competitive(cls) -> Self:
        return cls(0, 0)


@dataclass
class Status(Generic[TMove]):
    is_in_progress: bool
    played_by: int
    value: float
    legal_moves: list[TMove]

    def outcome(self, perspective: int) -> float:
        return self.value if perspective == self.played_by else -self.value



import matplotlib.pyplot as plt
import numpy as np


def create_checkered_array(n: int) -> np.ndarray:
    arr = np.ones((n, n), dtype=int)
    arr[1::2, ::2] = 0
    arr[::2, 1::2] = 0
    return arr


def checker_pattern_color(i, j):
    is_white = (i + j) % 2 == 0
    return "#FFFFFF" if is_white else "#000000"


def plot_state(planes: np.ndarray, figsize: tuple[int, int], linewidth: float = 1) -> None:
    _, rows, cols = planes.shape
    _, ax = plt.subplots(figsize=figsize)
    _ = ax.imshow(planes.transpose(1, 2, 0))

    ax.set_xticks(np.arange(-0.5, rows, 1), minor=True)
    ax.set_yticks(np.arange(-0.5, cols, 1), minor=True)
    ax.grid(which="minor", color="white", linestyle="-", linewidth=linewidth)

    ax.spines["top"].set_edgecolor("white")
    ax.spines["right"].set_edgecolor("white")
    ax.spines["bottom"].set_edgecolor("white")
    ax.spines["left"].set_edgecolor("white")

    ax.tick_params(axis="both", which="both", bottom=False, left=False)
    ax.set_xticks([])
    ax.set_yticks([])

    ax.set_title("Player   =   Red\nOpponent = Green", loc="right", fontname="Monospace", fontsize=9)


def plot_chess_slice(
    planes: np.ndarray, slice: int, vmax: float = 1.0, figsize: tuple[int, int] | None = None
) -> None:
    CHECK_DARKNESS = 0.1 * vmax
    _, rows, cols = planes.shape

    if slice < 0:
        plane = np.max(planes, axis=0)
    else:
        plane = planes[slice, :, :]
    plane += CHECK_DARKNESS * create_checkered_array(rows)

    figsize = figsize or (3, 3)
    _, ax = plt.subplots(figsize=figsize)

    cmap = plt.get_cmap("inferno")
    vmin, vmax = 0, vmax + CHECK_DARKNESS
    _ = ax.imshow(plane, cmap=cmap, vmin=vmin, vmax=vmax)

    ax.set_xticks(np.arange(-0.5, rows, 1), minor=True)
    ax.set_yticks(np.arange(-0.5, cols, 1), minor=True)
    ax.grid(which="minor", color="white", linestyle="-", linewidth=2.0)

    ax.spines["top"].set_edgecolor("white")
    ax.spines["right"].set_edgecolor("white")
    ax.spines["bottom"].set_edgecolor("white")
    ax.spines["left"].set_edgecolor("white")

    ax.tick_params(axis="both", which="both", bottom=False, left=False)
    ax.set_xticks([])
    ax.set_yticks([])


def heatmap_color(value: float) -> str:
    cmap = plt.get_cmap("summer")

    normalized_value = 1 - max(0, min(1, value))
    colormap_index = normalized_value * (cmap.N - 1)
    rgb_value = cmap(int(colormap_index))

    return "#{:02x}{:02x}{:02x}".format(
        int(rgb_value[0] * 255),
        int(rgb_value[1] * 255),
        int(rgb_value[2] * 255),
    )


def combine_hex_colors(color1, color2, transparency2):
    rgb1 = (int(color1[1:3], 16), int(color1[3:5], 16), int(color1[5:7], 16))
    rgb2 = (int(color2[1:3], 16), int(color2[3:5], 16), int(color2[5:7], 16))

    combined_color = [int(rgb2[i] * transparency2 + rgb1[i] * (1 - transparency2)) for i in range(3)]
    return "#{:02x}{:02x}{:02x}".format(*combined_color[:3])


def heatmap_checker_colors(matrix: np.ndarray, include_checkers: bool = True) -> np.ndarray:
    if not include_checkers:
        return np.vectorize(heatmap_color)(matrix)

    heatmap_colors = np.vectorize(heatmap_color)(matrix)
    checker_pattern_colors = np.vectorize(checker_pattern_color)(*np.indices(matrix.shape))
    combined_colors = np.vectorize(combine_hex_colors)(heatmap_colors, checker_pattern_colors, 0.08)
    return combined_colors


def light_checker_colors(shape: tuple[int, int]) -> np.ndarray:
    def col(row, col) -> str:
        return "#FFFFFF" if (row + col) % 2 == 1 else "#EBEBEB"

    return np.vectorize(col)(*np.indices(shape))



from copy import copy

from graphviz import Digraph

from .solvers.mcts import Node
from .states import State, TMove


def visualize_tree(state: State[TMove], root: Node) -> Digraph:
    dot = Digraph(comment="Game tree")

    def add_node(node: Node, state: State[TMove]):
        node_id = f"{id(node)}"

        # add current node
        dot.node(node_id, label=node.html(state), shape="rectangle")

        # add edges to children
        for child in node.children:
            child_id = f"{id(child)}"
            dot.edge(node_id, child_id)

            # recursively add nodes
            child_state = copy(state)
            child_state.play_move(child.move)
            add_node(child, child_state)

    # start with root node
    add_node(root, state)

    # render graph
    return dot



from .alpha_zero import AlphaZero as AlphaZero
from .alpha_zero_parameters import AZArenaParameters as AZArenaParameters
from .alpha_zero_parameters import AZMctsParameters as AZMctsParameters
from .alpha_zero_parameters import AZNetworkParameters as AZNetworkParameters
from .alpha_zero_parameters import AZTrainingParameters as AZTrainingParameters
from .network import NeuralNetwork as NeuralNetwork
from .network import TrainingData as TrainingData
from .network_torch import PytorchNeuralNetwork as PytorchNeuralNetwork
from .supervised_learning import SupervisedTrainer as SupervisedTrainer



from dataclasses import dataclass
from typing import Self

import chess
import numpy as np
from chess import Board, Move

from .chess_enums import ActionPlanes, ObsPlanes
from .state import Policy, State, Status

WHITE_IDXS = np.flipud(np.arange(64, dtype=np.uint64).reshape(8, 8))
WHITE_POWERS = 2**WHITE_IDXS

BLACK_IDXS = np.rot90(WHITE_IDXS, 2)
BLACK_POWERS = 2**BLACK_IDXS

PIECES = [chess.PAWN, chess.ROOK, chess.KNIGHT, chess.BISHOP, chess.QUEEN, chess.KING]


@dataclass(slots=True)
class ChessState(State[Move]):
    board: Board

    @property
    def rows(self) -> int:
        return 8

    @property
    def cols(self) -> int:
        return 8

    @property
    def shape(self) -> tuple[int, int]:
        return self.rows, self.cols

    @property
    def played_by(self) -> int:
        return 1 + self.board.turn

    @property
    def move_count(self) -> int:
        return len(self.board.move_stack)

    def status(self) -> Status[int]:
        board = self.board
        if self.board.is_checkmate():
            return Status(False, self.played_by, 1.0, [])

        if board.is_fifty_moves() or board.is_insufficient_material() or board.is_repetition():
            return Status(False, self.played_by, 0.0, [])

        if board.is_repetition():
            return Status(False, self.played_by, 0.0, [])

        # Stalemate
        legal_moves = list(self.board.legal_moves)
        has_legal_moves = bool(legal_moves)
        return Status(has_legal_moves, self.played_by, 0.0, legal_moves)

    def play_move(self, move: Move) -> Self:
        new_board = self.board.copy()
        new_board.push(move)
        return ChessState(new_board)

    def set_move(self, move: Move) -> None:
        self.board.push(move)

    def policy_loc(self, move: Move) -> tuple[int, int, int]:
        _, rows, cols = ActionPlanes.shape()
        p, r, c = self.policy_loc_3d(move)
        idx = p * (rows * cols) + r * cols + c
        return idx

    def policy_loc_3d(self, move: Move) -> tuple[int, int, int]:
        if self.board.turn:
            from_idx, to_idx = move.from_square, move.to_square
        else:
            from_idx, to_idx = 63 - move.from_square, 63 - move.to_square

        x_from, x_to = from_idx % 8, to_idx % 8
        y_from, y_to = from_idx // 8, to_idx // 8
        x_shift = x_to - x_from
        y_shift = y_to - y_from

        if move.promotion and move.promotion != chess.QUEEN:
            match x_shift, y_shift, move.promotion:
                case (-1, 1, chess.KNIGHT):
                    return ActionPlanes.PROMOTE_KNIGHT_NW, 7 - y_from, x_from
                case (0, 1, chess.KNIGHT):
                    return ActionPlanes.PROMOTE_KNIGHT_N, 7 - y_from, x_from
                case (1, 1, chess.KNIGHT):
                    return ActionPlanes.PROMOTE_KNIGHT_NE, 7 - y_from, x_from
                case (-1, 1, chess.ROOK):
                    return ActionPlanes.PROMOTE_ROOK_NW, 7 - y_from, x_from
                case (0, 1, chess.ROOK):
                    return ActionPlanes.PROMOTE_ROOK_N, 7 - y_from, x_from
                case (1, 1, chess.ROOK):
                    return ActionPlanes.PROMOTE_ROOK_NE, 7 - y_from, x_from
                case (-1, 1, chess.BISHOP):
                    return ActionPlanes.PROMOTE_BISHOP_NW, 7 - y_from, x_from
                case (0, 1, chess.BISHOP):
                    return ActionPlanes.PROMOTE_BISHOP_N, 7 - y_from, x_from
                case (1, 1, chess.BISHOP):
                    return ActionPlanes.PROMOTE_BISHOP_NE, 7 - y_from, x_from
                case _:
                    msg = f"x={x_shift}, y={y_shift}, promotion={move.promotion}"
                    raise ValueError(f"This should never happen ({msg})")

        # N, NE, E, SE, S, SW, W, NW
        match x_shift, y_shift:
            case (0, y):
                if y_shift > 0:
                    return 0 + 8 * (y_shift - 1), 7 - y_from, x_from
                else:
                    return 4 + 8 * (-y_shift - 1), 7 - y_from, x_from
            case (x, 0):
                if x_shift > 0:
                    return 2 + 8 * (x_shift - 1), 7 - y_from, x_from
                else:
                    return 6 + 8 * (-x_shift - 1), 7 - y_from, x_from
            case (x, y) if x == y:
                if x_shift > 0:
                    return 1 + 8 * (x_shift - 1), 7 - y_from, x_from
                else:
                    return 5 + 8 * (-x_shift - 1), 7 - y_from, x_from
            case (x, y) if x == -y:
                if x_shift > 0:
                    return 3 + 8 * (x_shift - 1), 7 - y_from, x_from
                else:
                    return 7 + 8 * (-x_shift - 1), 7 - y_from, x_from
            case (1, 2):
                return ActionPlanes.KNIGHT_NNE, 7 - y_from, x_from
            case (2, 1):
                return ActionPlanes.KNIGHT_ENE, 7 - y_from, x_from
            case (2, -1):
                return ActionPlanes.KNIGHT_ESE, 7 - y_from, x_from
            case (1, -2):
                return ActionPlanes.KNIGHT_SSE, 7 - y_from, x_from
            case (-1, -2):
                return ActionPlanes.KNIGHT_SSW, 7 - y_from, x_from
            case (-2, -1):
                return ActionPlanes.KNIGHT_WSW, 7 - y_from, x_from
            case (-2, 1):
                return ActionPlanes.KNIGHT_WNW, 7 - y_from, x_from
            case (-1, 2):
                return ActionPlanes.KNIGHT_NNW, 7 - y_from, x_from

        raise ValueError("This should never happen")

    def to_feature(self) -> np.ndarray:
        PIECE_COUNT = 6
        COL = 1
        CASTLING = 2
        NO_PROG = 1
        EN_PASSANT = 1
        TWOFOLD = 1

        PLANES = 2 * PIECE_COUNT + COL + 2 * CASTLING + NO_PROG + EN_PASSANT + TWOFOLD

        can_white_queen_castle = np.sign(self.board.castling_rights & chess.BB_A1)
        can_white_king_castle = np.sign(self.board.castling_rights & chess.BB_H1)
        can_black_queen_castle = np.sign(self.board.castling_rights & chess.BB_A8)
        can_black_king_castle = np.sign(self.board.castling_rights & chess.BB_H8)

        if self.board.turn == chess.WHITE:
            player_idxs = WHITE_IDXS
            player_powers = WHITE_POWERS
            can_player_queenside = can_white_queen_castle
            can_player_kingside = can_white_king_castle
            can_opponent_queenside = can_black_queen_castle
            can_opponent_kingside = can_black_king_castle
        else:
            player_idxs = BLACK_IDXS
            player_powers = BLACK_POWERS
            can_player_queenside = can_black_queen_castle
            can_player_kingside = can_black_king_castle
            can_opponent_queenside = can_white_queen_castle
            can_opponent_kingside = can_white_king_castle

        feature = np.zeros((PLANES, 8, 8), dtype=np.float32)
        for i, piece in enumerate(PIECES):
            player_pieces = self.board.pieces_mask(piece, self.board.turn)
            feature[i, :, :] = np.sign(player_powers & player_pieces)

        for i, piece in enumerate(PIECES):
            opponent_pieces = self.board.pieces_mask(piece, not self.board.turn)
            feature[PIECE_COUNT + i, :, :] = np.sign(player_powers & opponent_pieces)

        is_twofold = self.board.is_repetition(2)

        feature[ObsPlanes.TURN, :, :] = self.board.turn
        feature[ObsPlanes.CAN_PLAYER_KINGSIDE, :, :] = can_player_kingside
        feature[ObsPlanes.CAN_PLAYER_QUEENSIDE, :, :] = can_player_queenside
        feature[ObsPlanes.CAN_OPP_KINGSIDE, :, :] = can_opponent_kingside
        feature[ObsPlanes.CAN_OPP_QUEENSIDE, :, :] = can_opponent_queenside
        feature[ObsPlanes.HALFMOVE_CLOCK, :, :] = self.board.halfmove_clock
        feature[ObsPlanes.EN_PASSANT_SQ, :, :] = self.board.ep_square == player_idxs
        feature[ObsPlanes.IS_TWOFOLD, :, :] = is_twofold

        return feature

    def to_grid(self) -> np.ndarray:
        ...

    def get_input_move(self) -> Move:
        while True:
            legal_moves = self.status().legal_moves
            msg = "Your turn to move. Please enter a move (integer): "
            for i, move in enumerate(legal_moves):
                if i % 8 == 0:
                    msg += "\n"
                move_str = f"{i + 1: >2}. {move}"
                msg += f"{move_str: <10}"
            response = input(msg + "\n")

            try:
                idx = int(response)
                if 0 < idx and idx <= len(legal_moves):
                    return legal_moves[idx - 1]
            except ValueError:
                pass
            print("Move was invalid. Please try again.")

    def html(self, policy: np.ndarray | None = None, is_tiny_repr: bool = False) -> str:
        ...

    def plot(self, observation_plane: ObsPlanes) -> None:
        from ..views.plot import plot_chess_slice

        input_feature = self.to_feature()
        plot_chess_slice(input_feature, observation_plane)

    def plot_policy(self, action_plane: ActionPlanes, policy: Policy | None = None) -> None:
        from IPython.display import Math, display

        from ..views.plot import plot_chess_slice

        if policy is not None:
            value_expression = rf"v \in (-1, +1) = {policy.value:.3f}"
            display(Math(value_expression))

            encoded_policy = policy.encoded_policy
            vmax = encoded_policy.max()
            policy = encoded_policy.reshape(ActionPlanes.shape())
            plot_chess_slice(policy, action_plane, vmax=vmax)
        else:
            encoded_policy = np.zeros(ActionPlanes.shape())
            for move in self.status().legal_moves:
                p, r, c = self.policy_loc_3d(move)
                encoded_policy[p, r, c] = 1.0

            plot_chess_slice(encoded_policy, action_plane)

    def csv(self) -> str:
        sans = []
        new_board = chess.Board()
        for move in self.board.move_stack:
            san = new_board.san(move)
            new_board.push(move)
            sans.append(san)

        return f"[{','.join(sans)}]"

    def pgn(self, title: str | None = None) -> str:
        from datetime import datetime

        import chess.pgn

        date = datetime.now()
        formatted_date = date.strftime("%Y.%m.%d")

        pgn = chess.pgn.Game()
        pgn.headers["Event"] = title or "Example"
        pgn.headers["Date"] = formatted_date

        if self.board.move_stack:
            node = pgn.add_variation(self.board.move_stack[0])

        for move in self.board.move_stack[1:]:
            node = node.add_variation(move)

        return str(pgn)

    def __copy__(self) -> "ChessState":
        board = self.board.copy()
        return ChessState(board)

    @classmethod
    def create(cls, fen: str | list[str] | None = None) -> "ChessState":
        board = Board()
        if fen:
            if isinstance(fen, str):
                board.set_fen(fen)
            elif isinstance(fen, list):
                for san in fen:
                    board.push_san(san)

        return ChessState(board)



from enum import Enum


class GameType(Enum):
    """Enum representing the types of games supported."""

    TICTACTOE = "TICTACTOE"
    CONNECTX = "CONNECTX"
    CHESS = "CHESS"

    @staticmethod
    def from_str(value: str) -> "GameType":
        """Converts a string to its enum representation.

        Args:
            value (str): The string representation of the game type.

        Raises:
            ValueError: If the string is not recognised.

        Returns:
            GameType: The enum
        """
        if value.upper().startswith("TIC"):
            return GameType.TICTACTOE
        if value.upper() == "CONNECTX":
            return GameType.CONNECTX
        if value.upper() == "CHESS":
            return GameType.CHESS
        supported_types = ", ".join([e.name for e in GameType])
        message = f"{value} not a supported solver type. Supported types are {supported_types}."
        raise ValueError(message)



import pickle
from copy import deepcopy
from pathlib import Path
from typing import Self

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from numpy.typing import NDArray
from torch.optim import Adam, Optimizer
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.tensorboard import SummaryWriter
from tqdm import trange

from ..games import Game
from ..states import State, TMove
from .alpha_zero_parameters import AZNetworkParameters
from .network import TrainingData

VALUE_FILTERS = 1
POLICY_FILTERS = 2
HEAD_KERNEL_SIZE = 1
KERNEL_SIZE = 3
PADDING = (KERNEL_SIZE - 1) // 2


class ResNet(nn.Module):
    def __init__(
        self,
        observation_shape: tuple[int, int, int],
        action_size: int,
        num_res_blocks: int = 4,
        num_features: int = 64,
    ) -> None:
        super().__init__()

        num_planes, rows, cols = observation_shape

        # We start with a convolutional layer
        self.start_block = nn.Sequential(
            nn.Conv2d(num_planes, num_features, kernel_size=KERNEL_SIZE, padding=PADDING),
            nn.BatchNorm2d(num_features),
            nn.ReLU(inplace=True),
        )

        # This feeds into 'num_res_blocks' residual layers
        self.res_blocks = nn.ModuleList([ResNetBlock(num_features) for _ in range(num_res_blocks)])

        # The value head derives a value, v, for the current position. v ∈ [-1, +1]
        self.value_head = nn.Sequential(
            nn.Conv2d(num_features, VALUE_FILTERS, kernel_size=HEAD_KERNEL_SIZE),
            nn.BatchNorm2d(VALUE_FILTERS),
            nn.ReLU(inplace=True),
            nn.Flatten(),
            nn.Linear(VALUE_FILTERS * rows * cols, num_features),
            nn.ReLU(inplace=True),
            nn.Linear(num_features, 1),
            nn.Tanh(),
        )

        self.policy_head = nn.Sequential(
            nn.Conv2d(num_features, POLICY_FILTERS, kernel_size=HEAD_KERNEL_SIZE),
            nn.BatchNorm2d(POLICY_FILTERS),
            nn.ReLU(inplace=True),
            nn.Flatten(),
            nn.Linear(POLICY_FILTERS * rows * cols, action_size),
        )

    def forward(self, x):
        output = self.start_block(x)
        for res_block in self.res_blocks:
            output = res_block(output)

        policy = self.policy_head(output)
        value = self.value_head(output)

        return policy, value


class ResNetBlock(nn.Module):
    def __init__(self, num_features: int):
        super().__init__()

        self.layers = nn.Sequential(
            nn.Conv2d(num_features, num_features, kernel_size=KERNEL_SIZE, padding=PADDING),
            nn.BatchNorm2d(num_features),
            nn.ReLU(inplace=True),
            nn.Conv2d(num_features, num_features, kernel_size=KERNEL_SIZE, padding=PADDING),
            nn.BatchNorm2d(num_features),
        )

        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        output = self.layers(x)
        output += x
        return self.relu(output)


class PytorchNeuralNetwork:
    def __init__(
        self,
        resnet: ResNet,
        optimizer: Optimizer,
        game: Game,
        generation: int,
        directory: Path,
    ) -> None:
        self.resnet = resnet
        self.optimizer = optimizer
        self.game = game
        self.directory = directory
        self.generation = generation
        self.device = next(resnet.parameters()).device.type

    @torch.no_grad()
    def predict(self, state: State[TMove]) -> tuple[NDArray[np.float32], float]:
        # Convert to [player, opponent, unplayed]
        planes = state.to_feature()
        tensor = torch.tensor(planes, device=self.device).unsqueeze(0)

        predicted_policy, predicted_outcome = self.resnet(tensor)
        normalized_policy = F.softmax(predicted_policy.squeeze(), dim=0).cpu().numpy()

        return normalized_policy, predicted_outcome.item()

    @torch.no_grad()
    def predict_parallel(self, states: list[State]) -> tuple[NDArray[np.float32], NDArray[np.float32]]:
        encoded_states = np.array([s.to_feature() for s in states], dtype=np.float32)
        torch_states = torch.tensor(encoded_states, device=self.device)

        predicted_policies, predicted_outcomes = self.resnet(torch_states)
        normalized_policies = torch.softmax(predicted_policies, axis=1).cpu().numpy()

        return normalized_policies, predicted_outcomes.squeeze().cpu().numpy()

    def set_to_eval(self) -> None:
        self.resnet.eval()

    def train(
        self,
        training_set: list[TrainingData],
        num_epochs: int,
        minibatch_size: int,
        log_progress: bool = True,
    ) -> None:
        # Initialize TensorBoard writer
        writer = SummaryWriter(log_dir=f"runs/{self.game.fullname}")
        gen = self.generation
        num_points = len(training_set)

        # Convert to numpy arrays for training
        states_np = np.array([d.encoded_state for d in training_set], dtype=np.float32)
        policies_np = np.array([d.policy for d in training_set], dtype=np.float32)
        outcomes_np = np.array([d.outcome for d in training_set], dtype=np.float32).reshape(-1, 1)

        states = torch.tensor(states_np, device=self.device)
        policies = torch.tensor(policies_np, device=self.device)
        outcomes = torch.tensor(outcomes_np, device=self.device)

        data_set = TensorDataset(states, policies, outcomes)
        data_loader = DataLoader(data_set, batch_size=minibatch_size, shuffle=True)

        if log_progress and gen == 0:
            with torch.no_grad():
                batch_states, batch_policies, batch_outcomes = next(iter(data_loader))
                grid = torchvision.utils.make_grid(batch_states)
                writer.add_graph(self.resnet, batch_states)
                if batch_states.shape[1] == 3:
                    writer.add_image("images", grid, 0)

            writer.close()

        # Record how well the previous generation predicts the next set of policies
        if log_progress:
            self.resnet.eval()
            epoch_policy_loss = 0.0
            epoch_outcome_loss = 0.0
            epoch_total_loss = 0.0
            with torch.no_grad():
                for batch_states, batch_policies, batch_outcomes in data_loader:
                    predicted_logits, predicted_outcomes = self.resnet(batch_states)

                    # Compute loss
                    policy_loss = F.cross_entropy(predicted_logits, batch_policies)
                    outcome_loss = F.mse_loss(predicted_outcomes, batch_outcomes)
                    total_loss = policy_loss + outcome_loss

                    # Store key metrics
                    epoch_policy_loss += policy_loss.item()
                    epoch_outcome_loss += outcome_loss.item()
                    epoch_total_loss += total_loss.item()

            # Logging metrics to TensorBoard
            writer.add_scalar(f"Generations/Total Loss", epoch_total_loss / num_points, gen)
            writer.add_scalar(f"Generations/Policy Loss", epoch_policy_loss / num_points, gen)
            writer.add_scalar(f"Generations/Outcome Loss", epoch_outcome_loss / num_points, gen)

        self.resnet.train()
        for epoch in trange(num_epochs, desc=" - Training", leave=False):
            epoch_policy_loss = 0.0
            epoch_outcome_loss = 0.0
            epoch_total_loss = 0.0

            for batch_states, batch_policies, batch_outcomes in data_loader:
                # zero the parameter gradients
                self.optimizer.zero_grad()

                # Feed forwards
                predicted_logits, predicted_outcomes = self.resnet(batch_states)

                # Compute loss
                policy_loss = F.cross_entropy(predicted_logits, batch_policies)
                outcome_loss = 5 * F.mse_loss(predicted_outcomes, batch_outcomes)
                total_loss = policy_loss + outcome_loss

                # Backwards + optimize
                total_loss.backward()
                self.optimizer.step()

                # Store key metrics
                epoch_policy_loss += policy_loss.item()
                epoch_outcome_loss += outcome_loss.item()
                epoch_total_loss += total_loss.item()

            # Logging metrics to TensorBoard
            if log_progress:
                writer.add_scalar(f"Gen{gen:02d}/Total Loss", epoch_total_loss, epoch + 1)
                writer.add_scalar(f"Gen{gen:02d}/Policy Loss", epoch_policy_loss, epoch + 1)
                writer.add_scalar(f"Gen{gen:02d}/Outcome Loss", epoch_outcome_loss, epoch + 1)

        # Close TensorBoard writer
        writer.close()

    def save_training_data(self, training_set: list[TrainingData]) -> None:
        gen = self.generation
        data_directory = self.directory / "training_data"
        data_file = data_directory / f"{self.game.fullname}_gen{gen:03d}.pkl"

        with open(data_file, "wb") as f:
            pickle.dump(training_set, f)

    def load_training_data(self, gen: int) -> list[TrainingData]:
        data_directory = self.directory / "training_data"
        data_file = data_directory / f"{self.game.fullname}_gen{gen:03d}.pkl"

        with open(data_file, "rb") as f:
            return pickle.load(f)

    def save(self) -> None:
        gen = self.generation
        weights_directory = self.directory / "weights"
        network_weights = self.resnet.state_dict()
        optimizer_state = self.optimizer.state_dict()
        weights_file = weights_directory / f"weights_{self.game.fullname}_gen{gen:03d}.pt"
        optimizer_file = weights_directory / f"optimizer_state_{self.game.fullname}_gen{gen:03d}.pt"
        torch.save(network_weights, weights_file)
        torch.save(optimizer_state, optimizer_file)

    def __deepcopy__(self, memo) -> Self:
        # Deep copy the model and optimizer
        resnet = deepcopy(self.resnet)
        optimizer = type(self.optimizer)(resnet.parameters(), **self.optimizer.defaults)

        resnet.load_state_dict(self.resnet.state_dict())
        optimizer.load_state_dict(self.optimizer.state_dict())

        game = deepcopy(self.game)
        directory = deepcopy(self.directory)

        return PytorchNeuralNetwork(resnet, optimizer, game, self.generation, directory)

    @classmethod
    def create(cls, game: Game, root_directory: str | Path, load_latest: bool | int = True) -> Self:
        root_directory = Path(root_directory)
        directory = root_directory / "weights"
        game_parameters = game.config()
        obs_shape = game_parameters.observation_shape
        action_size = game_parameters.action_size

        device = "cuda" if torch.cuda.is_available() else "cpu"
        map_location = torch.device(device)

        net_params = AZNetworkParameters.defaults(game.fullname)
        resnet = ResNet(obs_shape, action_size, net_params.num_resnet_blocks, net_params.num_features)
        resnet = resnet.to(device)

        learning_rate = net_params.optimizer_learn_rate
        l2_regularization = net_params.optimizer_weight_decay
        optimizer = Adam(resnet.parameters(), lr=learning_rate, weight_decay=l2_regularization)

        weights_regex = f"weights_{game.fullname}_gen*.pt"
        optimizer_regex = f"optimizer_state_{game.fullname}_gen*.pt"
        weights_files = [file_path for file_path in directory.glob(weights_regex)]
        optim_files = [file_path for file_path in directory.glob(optimizer_regex)]

        generation = 0
        if load_latest:
            if not isinstance(load_latest, bool):
                weights_files = [f for f in weights_files if int(f.stem.split("gen")[1]) == load_latest]
                optim_files = [f for f in optim_files if int(f.stem.split("gen")[1]) == load_latest]

            if weights_files:
                latest_weights_file = max(weights_files, key=lambda f: int(f.stem.split("gen")[1]))
                resnet.load_state_dict(torch.load(latest_weights_file, map_location=map_location))
                generation = int(latest_weights_file.stem.split("gen")[1])

            if optim_files:
                latest_optimiser_file = max(optim_files, key=lambda f: int(f.stem.split("gen")[1]))
                optimizer.load_state_dict(torch.load(latest_optimiser_file, map_location=map_location))

        return PytorchNeuralNetwork(resnet, optimizer, game, generation, root_directory)



from abc import ABC, abstractmethod
from typing import Generic

from ..states import TState


class View(ABC, Generic[TState]):
    @abstractmethod
    def display(self, state: TState) -> None:
        ...
